unit CompStartup;

{
  Inno Setup
  Copyright (C) 1998-2001 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Compiler Startup form

  $jrsoftware: issrc/Projects/CompStartup.pas,v 1.6 2002/11/09 19:45:41 jr Exp $
}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type

  TStartupFormResult = (srNone, srEmpty, srWizard, srOpenFile, srOpenDialog);

  TStartupForm = class(TForm)
    OKButton: TButton;
    CancelButton: TButton;
    GroupBox1: TGroupBox;
    GroupBox2: TGroupBox;
    EmptyRadioButton: TRadioButton;
    WizardRadioButton: TRadioButton;
    OpenRadioButton: TRadioButton;
    OpenListBox: TListBox;
    StartupCheck: TCheckBox;
    NewImage: TImage;
    OpenImage: TImage;
    procedure RadioButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure DblClick_(Sender: TObject);
    procedure OKButtonClick(Sender: TObject);
  private
    FResult: TStartupFormResult;
    FResultFileName: TFileName;
    procedure SetMRUList(const MRUList: TStringList);
    procedure MakeTransparent(const Source: TBitmap; const Color: TColor);
  public
    property MRUList: TStringList write SetMRUList;
    property Result: TStartupFormResult read FResult;
    property ResultFileName: TFileName read FResultFileName;
  end;

implementation

uses
  CompMsgs, CmnFunc, CmnFunc2, CompForm;

{$R *.DFM}

procedure TStartupForm.SetMRUList(const MRUList: TStringList);
var
  I: Integer;
begin
  for I := 0 to MRUList.Count-1 do
    OpenListBox.Items.Add(MRUList[I]);
  UpdateHorizontalExtent(OpenListBox);
end;

procedure TStartupForm.MakeTransparent(const Source: TBitmap; const Color: TColor);
var
  B: TBitmap;
  R: TRect;
begin
  Source.PixelFormat := pfDevice;
  B := TBitmap.Create;
  try
    R := Rect(0, 0, Source.Width, Source.Height);
    B.Width := Source.Width;
    B.Height := Source.Height;
    B.Canvas.Brush.Color := clBtnFace;
    B.Canvas.BrushCopy(R, Source, R, Color);
    Source.Canvas.CopyRect(R, B.Canvas, R);
  finally
    B.Free;
  end;
end;

procedure TStartupForm.FormCreate(Sender: TObject);
begin
  FResult := srNone;

  InitFormFont(Self);

  MakeTransparent(NewImage.Picture.Bitmap, clSilver);
  MakeTransparent(OpenImage.Picture.Bitmap, clSilver);

  OpenListBox.Items.Add(SCompilerMoreFiles);
  OpenListBox.ItemIndex := 0;
  UpdateHorizontalExtent(OpenListBox);
  ActiveControl := OpenRadioButton;
end;

procedure TStartupForm.RadioButtonClick(Sender: TObject);
begin
  EmptyRadioButton.Checked := Sender = EmptyRadioButton;
  WizardRadioButton.Checked := Sender = WizardRadioButton;
  OpenRadioButton.Checked := Sender = OpenRadioButton;
  OpenListBox.Enabled := Sender = OpenRadioButton;
end;

procedure TStartupForm.DblClick_(Sender: TObject);
begin
  if OkButton.Enabled then
    OkButton.Click;
end;

procedure TStartupForm.OKButtonClick(Sender: TObject);
begin
  if EmptyRadioButton.Checked then
    FResult := srEmpty
  else if WizardRadioButton.Checked then
    FResult := srWizard
  else { if OpenRadioButton.Checked then } begin
    if OpenListBox.ItemIndex > 0 then begin
      FResult := srOpenFile;
      FResultFileName := OpenListBox.Items[OpenListBox.ItemIndex];
    end else
      FResult := srOpenDialog;
  end;
end;

end.
