unit DebugClient;

{
  Inno Setup
  Copyright (C) 1998-2002 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Debug info stuff

  $jrsoftware: issrc/Projects/DebugClient.pas,v 1.3 2002/11/09 19:45:41 jr Exp $
}

interface

uses
  Windows, SysUtils, Messages, DebugStruct;

var
  Debugging: Boolean;

procedure DebugNotify(Kind: TDebugEntryKind; Index: Integer);
procedure EndDebug;
procedure SetDebugWnd(Wnd: HWND);

implementation

uses
  Forms, Main;

type
  TDummyClass = class
  private
    class procedure DebugClientWndProc(var Message: TMessage);
  end;

var
  DebugWnd: HWND;
  DebugClientWnd: HWND;
  DebugContinue: Boolean;

procedure SetDebugWnd(Wnd: HWND);
begin
  Debugging := True;
  DebugWnd := Wnd;
  AttachThreadInput(GetCurrentThreadId, GetWindowThreadProcessId(DebugWnd, nil),
    True);
  DebugClientWnd := AllocateHWnd(TDummyClass.DebugClientWndProc);
  SendMessage(DebugWnd, WM_Debugger_Hello, WPARAM(DebugClientWnd), 0);
end;

procedure EndDebug;
begin
  Debugging := False;
  if DebugWnd <> 0 then begin
    SendMessage(DebugWnd, WM_Debugger_Goodbye, 0, 0);
    DebugWnd := 0;
  end;
  if DebugClientWnd <> 0 then begin
    DeallocateHWnd(DebugClientWnd);
    DebugClientWnd := 0;
  end;
end;

procedure DebugNotify(Kind: TDebugEntryKind; Index: Integer);
var
  SaveAppTitle: String;
  WindowList: Pointer;
  Msg: TMsg;
begin
  if not Debugging then
    Exit;

  DebugContinue := False;
  if SendMessage(DebugWnd, WM_Debugger_Stepped, Ord(Kind), Index) = 0 then begin
    { Don't pause }
    Exit;
  end;

  { Wait until we get clearance to continue }
  SaveAppTitle := Application.Title;
  WindowList := DisableTaskWindows(0);
  try
    Application.Title := '[Paused] ' + SaveAppTitle;
    while not DebugContinue do begin
      case Integer(GetMessage(Msg, 0, 0, 0)) of
        -1: Break; { if GetMessage failed }
        0: begin
             { Repost WM_QUIT messages }
             PostQuitMessage (Msg.WParam);
             Break;
           end;
      end;
      TranslateMessage(Msg);
      DispatchMessage(Msg);
    end;
  finally
    EnableTaskWindows(WindowList);
    Application.Title := SaveAppTitle;
  end;
  { Bring us back to the foreground }
  Application.BringToFront;
end;

class procedure TDummyClass.DebugClientWndProc(var Message: TMessage);
var
  EvaluateExp, EvaluateResult: String;
begin
  try
    case Message.Msg of
      WM_DebugClient_Detach: begin
          Debugging := False;
          DebugWnd := 0;
          { If it's paused, force it to continue }
          DebugContinue := True;
          { Make the GetMessage call in DebugNotify return immediately }
          PostMessage(0, 0, 0, 0);
        end;
      WM_DebugClient_Continue: begin
          DebugContinue := True;
          { Make the GetMessage call in DebugNotify return immediately }
          PostMessage(0, 0, 0, 0);
        end;
      WM_COPYDATA: begin
          case TWMCopyData(Message).CopyDataStruct.dwData of
            CD_DebugClient_EvaluateConstant: begin
                try
                  SetString(EvaluateExp, PChar(TWMCopyData(Message).CopyDataStruct.lpData),
                    TWMCopyData(Message).CopyDataStruct.cbData);
                  try
                    ConstReadOnly := True;
                    try
                      EvaluateResult := ChangeDirConst(EvaluateExp);
                    finally
                      ConstReadOnly := False;
                    end;
                    ReplyMessage(1);  { send back 'success' reply now }
                  except
                    on E: Exception do begin
                      EvaluateResult := E.Message;
                      ReplyMessage(2);
                    end;
                  end;
                  SendCopyDataMessage(DebugWnd, DebugClientWnd, CD_Debugger_Reply,
                    Pointer(EvaluateResult), Length(EvaluateResult));
                except
                  { don't propogate exceptions }
                end;
              end;
          end;
        end;
    else
      with Message do
        Result := DefWindowProc(DebugClientWnd, Msg, WParam, LParam);
    end;
  except
    Application.HandleException(nil);
  end
end;

end.
