program ISCC;
{$APPTYPE CONSOLE}

{
  Inno Setup
  Copyright (C) 1998-2002 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Command-line compiler

  $jrsoftware: issrc/Projects/ISCC.dpr,v 1.9 2002/11/09 19:45:41 jr Exp $
}

uses
  Windows, SysUtils, MBCSUtil, CmnFunc2, CompInt;

{$R *.RES}

{$I VERSION.INC}

var
  F: TextFile;
  CurLine: String;
  StartTime, EndTime: DWORD;
  WantAbort: Boolean;

function ConsoleCtrlHandler (dwCtrlType: DWORD): BOOL; stdcall;
begin
  { Abort gracefully when Ctrl+C/Break is pressed }
  WantAbort := True;
  Result := True;
end;

{$IFNDEF JR_D3}
{ Delphi 2's internal TextIn function doesn't handle the ERROR_BROKEN_PIPE
  error properly. Work around this. }
procedure NewTextIn;
asm
  { Call previous handler }
  call dword ptr [eax].TTextRec.UserData
  { If result is ERROR_BROKEN_PIPE, change it to zero }
  cmp  eax, ERROR_BROKEN_PIPE
  jne  @@exit
  xor  eax, eax
  @@exit:
end;
{$ENDIF}

function CompilerCallbackProc (Code: Integer; var Data: TCompilerCallbackData;
  AppData: Longint): Integer; stdcall;
begin
  if WantAbort then begin
    Result := iscrRequestAbort;
    Exit;
  end;
  Result := iscrSuccess;
  case Code of
    iscbReadScript: begin
        { Note: In Inno Setup 3.0.1 and later we can ignore Data.Reset since
          it is only True once (when reading the first line). }
        if not Eof(F) then begin
          Readln (F, CurLine);
          Data.LineRead := PChar(CurLine);
        end;
      end;
    iscbNotifyStatus:
      Writeln (Data.StatusMsg);
    iscbNotifySuccess: begin
        EndTime := GetTickCount;
        Writeln;
        Writeln ('Successful compile (', (EndTime - StartTime) / 1000 : 0 : 3,
          ' sec). Resulting Setup program filename is:');
        Writeln (Data.OutputExeFilename);
      end;
    iscbNotifyError:
      if Assigned(Data.ErrorMsg) then begin
        Write ('Error');
        if Data.ErrorLine <> 0 then
          Write (' on line ', Data.ErrorLine);
        if Assigned(Data.ErrorFilename) then
          Write (' in ', Data.ErrorFilename);
        Writeln (': ', Data.ErrorMsg);
      end;
  end;
end;

procedure Go;
var
  ScriptFilename, ScriptPath: String;
  ExitCode: Integer;
  Ver: PCompilerVersionInfo;
  Params: TCompileScriptParams;
begin
  Writeln ('Inno Setup 3 Command-Line Compiler (b)');
  Writeln ('Copyright (C) 1998-2002 Jordan Russell. All rights reserved.');
  Writeln ('Portions by Martijn Laan');
  Writeln;
  if (ParamCount <> 1) or (ParamStr(1) = '/?') then begin
    Writeln ('Usage:  iscc scriptfile.iss');
    Writeln ('or to read from standard input:  iscc -');
    Halt (1);
  end;

  if ParamStr(1) <> '-' then begin
    ScriptFilename := ExpandFilename(ParamStr(1));
    ScriptPath := ExtractFilePath(ScriptFilename);
  end
  else begin
    { Read from standard input }
    ScriptFilename := '';
    ScriptPath := GetCurrentDir;
  end;

  Ver := ISDllGetVersion;
  if Ver.BinVersion < $3000100 then begin
    Writeln ('Incompatible compiler engine version.');
    Halt (1);
  end;

  AssignFile (F, ScriptFilename);
  FileMode := fmOpenRead or fmShareDenyWrite;
  Reset (F);
  {$IFNDEF JR_D3}
  Pointer((@TTextRec(F).UserData)^) := TTextRec(F).InOutFunc;
  TTextRec(F).InOutFunc := @NewTextIn;
  {$ENDIF}
  ExitCode := 0;
  try
    Writeln ('Compiler engine version: ', Ver.Title, ' ', Ver.Version);
    Writeln;

    FillChar (Params, SizeOf(Params), 0);
    Params.Size := SizeOf(Params);
    Params.SourcePath := PChar(ScriptPath);
    Params.CallbackProc := CompilerCallbackProc;

    StartTime := GetTickCount;
    case ISDllCompileScript(Params) of
      isceInvalidParam: begin
          ExitCode := 1;
          Writeln ('Internal error 1.');
        end;
      isceCompileFailure: begin
          ExitCode := 2;
          Writeln ('Compile aborted.');
        end;
    end;
  finally
    CloseFile (F);
  end;
  if ExitCode <> 0 then
    Halt (ExitCode);
end;

procedure ShowExceptionMsg (Msg: String);
begin
  if (Msg <> '') and (AnsiLastChar(Msg)^ > '.') then Msg := Msg + '.';
  Writeln (Msg);
end;

begin
  SetConsoleCtrlHandler (@ConsoleCtrlHandler, True);
  try
    Go;
  except
    on E: Exception do begin
      { Show a friendlier exception message. (By default, Delphi prints out
        the exception class and address.) }
      ShowExceptionMsg (E.Message);
      Halt (2);
    end;
  end;
end.
