unit Main;

{
  Inno Setup
  Copyright (C) 1998-2002 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Background form

  $jrsoftware: issrc/Projects/Main.pas,v 1.49.4.1 2003/04/14 20:00:19 jr Exp $
}

interface

{$I VERSION.INC}

uses
  Windows, SysUtils, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Struct, DebugStruct;

const
  WM_SETUPNEXTSTEP = WM_USER + 3900;

type
  TCurStep = (csStart, csWizard, csCopy, csFinished, csTerminate);
  TMainForm = class(TForm)
    procedure FormResize(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormPaint(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    procedure AppOnActivate (Sender: TObject);
    function MainWindowHook (var Message: TMessage): Boolean;
    procedure WMSysCommand (var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure WMSetupNextStep (var Msg: TMessage); message WM_SETUPNEXTSTEP;
    procedure WMEraseBkgnd (var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMGetDlgCode (var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMShowWindow (var Message: TWMShowWindow); message WM_SHOWWINDOW;
  protected
    function GetPalette: HPALETTE; override;
  public
    { Public declarations }
    CurStep: TCurStep;
    IsMinimized, InExecuteLoop, HideWizard: Boolean;
    NewPalette: HPALETTE;
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CenterForm (AForm: TForm);
    procedure ProceedToNextStep;
    procedure ShowAboutBox;
    procedure UpdateWizardFormVisibility;
  end;

  TInstallOnThisVersionResult = (irInstall, irNotOnThisPlatform,
    irVerTooLow, irVerTooHigh);
  TShellFolderID = (sfDesktop, sfStartMenu, sfPrograms, sfStartup, sfSendTo,
    sfFonts, sfAppData, sfDocs, sfTemplates, sfFavorites, sfLocalAppData);
  TEntryType = (seType, seComponent, seTask, seDir, seFile, seFileLocation,
    seIcon, seIni, seRegistry, seInstallDelete, seUninstallDelete, seRun,
    seUninstallRun);

const
  EntryStrings: array[TEntryType] of Integer = (SetupTypeEntryStrings,
    SetupComponentEntryStrings, SetupTaskEntryStrings, SetupDirEntryStrings,
    SetupFileEntryStrings, SetupFileLocationEntryStrings, SetupIconEntryStrings,
    SetupIniEntryStrings, SetupRegistryEntryStrings, SetupDeleteEntryStrings,
    SetupDeleteEntryStrings, SetupRunEntryStrings, SetupRunEntryStrings);

  { Exit codes that are assigned to the SetupExitCode variable.
    Note: SetupLdr also returns exit codes with the same numbers. }
  ecInitializationError = 1;     { Setup failed to initialize. }
  ecCancelledBeforeInstall = 2;  { User clicked Cancel before the actual
                                   installation started. }
  ecNextStepError = 3;           { A fatal exception occurred while moving to
                                   the next step. }
  ecInstallationError = 4;       { A fatal exception occurred during
                                   installation. }
  ecInstallationCancelled = 5;   { User clicked Cancel during installation,
                                   or clicked Abort at an Abort-Retry-Ignore
                                   dialog. }

var
  MainForm: TMainForm;

  { Variables for command line parameters }
  SetupLdrMode: Boolean;
  SetupLdrWnd: HWND;
  SetupLdrOriginalFilename: String;
  SetupLdrOffsetMsg, SetupLdrOffset0, SetupLdrOffset1: Longint;
  InitDir, InitProgramGroup: String;
  InitLoadInf, InitSaveInf: String;
  InitNoIcons, InitSilent, InitVerySilent, InitNoRestart, InitNoCancel: Boolean;
  InitComponents: TStringList;
  DetachedUninstMsgFile: Boolean;

  { Debugger }
  OriginalEntryIndexes: array[TEntryType] of TList;

  { 'Constants' }
  SourceDir, TempInstallDir, WinDir, WinSystemDir, SystemDrive, FontDir,
    ProgramFilesDir, CommonFilesDir, DAODir, CmdFilename, SysUserInfoName,
    SysUserInfoOrg, UninstallExeFilename: String;

  { Variables read in from the SETUP.0 file }
  SetupHeader: TSetupHeader;
  LangOptions: TSetupLangOptions;
  Entries: array[TEntryType] of TList;
  WizardImage: TBitmap;
  WizardSmallImage: TBitmap;
  BunzipDLL: TMemoryStream;

  { User options }
  WizardUserInfoName, WizardUserInfoOrg, WizardDirValue, WizardGroupValue: String;
  WizardNoIcons: Boolean;
  WizardSetupType: PSetupTypeEntry;
  WizardComponents, WizardTasks: TStringList;
  NeedToAbortInstall: Boolean;

  { Other }
  HasIcons, IsNT, NewGUI, IsAdmin, IsPowerUser, NeedPassword, NeedsRestart,
    RestartSystem: Boolean;
  HasCustomType, HasComponents, HasTasks, HasTaskIcons: Boolean;
  RunCheckBoxes: TList;
  WindowsVersion: Cardinal;
  NTServicePackLevel: Word;
  MinimumSpace: Longint;
  DeleteFilesAfterInstallList, DeleteDirsAfterInstallList: TStringList;
  UninstallRegKeyBaseName: String;
  StartParam: Integer;
  ConstReadOnly: Boolean;
  SetupExitCode: Integer;

function ShouldProcessEntry(const WizardComponents, WizardTasks: TStringList;
  const Components, Tasks: String): Boolean;
function ShouldProcessRunEntry(const WizardComponents, WizardTasks: TStringList;
  const RunEntry: PSetupRunEntry): Boolean;
function ShouldProcessIconEntry(const WizardComponents, WizardTasks: TStringList;
  const WizardNoIcons: Boolean; const IconEntry: PSetupIconEntry): Boolean;
function ChangeDirConst (const S: String): String;
function ChangeDirConstEx (const S: String; const CustomConsts: array of String): String;
procedure DebugNotifyEntry(EntryType: TEntryType; Number: Integer);
procedure DeinitSetup;
function ExitSetupMsgBox: Boolean;
function GetDefaultDialogFontName: String;
function GetShellFolder (Common: Boolean; const ID: TShellFolderID;
  ReadOnly: Boolean): String;
procedure InitializeSetup;
function InstallOnThisVersion (const MinVersion: TSetupVersionData;
  const OnlyBelowVersion: TSetupVersionData): TInstallOnThisVersionResult;
function SetFontNameSize (const AFont: TFont; const AName: String;
  const ASize: Integer; const ADefaultName: String;
  const ADefaultSize: Integer): Boolean;
procedure InternalError (const Id: String);
procedure SetFormFont (Form: TForm; var OldTextHeight, NewTextHeight: Integer);

implementation

uses
  ShellAPI, ShlObj,
  Msgs, MsgIDs, ChildFrm, Install, InstFunc, InstFnc2, MBCSUtil, CmnFunc, CmnFunc2,
  zlib, bzlib, SetupEnt, Wizard, DebugClient, VerInfo;

{$R *.DFM}

var
  ShellFolders: array[Boolean, TShellFolderID] of String;
  ShellFoldersRead: array[Boolean, TShellFolderID] of Boolean;
  SHFolderDLLHandle, SHFolderShell32Handle: HMODULE;
  SHGetFolderPathFunc: function(hwndOwner: HWND; nFolder: Integer;
    hToken: THandle; dwFlags: DWORD; pszPath: PAnsiChar): HRESULT; stdcall;

  BunzipDLLHandle: HMODULE;

{ Misc. functions }

function ShouldProcessEntry(const WizardComponents, WizardTasks: TStringList;
  const Components, Tasks: String): Boolean;

  function ShouldProcessEntry(const WizardItems, EntryItems: TStringList): Boolean;
  var
    I: Integer;
  begin
    Result := False;
    for I := 0 to WizardItems.Count-1 do begin
      if EntryItems.IndexOf(WizardItems[I]) <> -1 then begin
        Result := True;
        Break;
      end;
    end;
  end;

var
  ProcessComponent, ProcessTask: Boolean;
  EntryItems: TStringList;
begin
  if (Components <> '') or (Tasks <> '') then begin
    EntryItems := TStringList.Create();

    if (Components <> '') and (WizardComponents <> nil) then begin
      EntryItems.CommaText := Components;
      ProcessComponent := ShouldProcessEntry(WizardComponents, EntryItems);
    end else
      ProcessComponent := True;

    if (Tasks <> '') and (WizardTasks <> nil) then begin
      EntryItems.CommaText := Tasks;
      ProcessTask := ShouldProcessEntry(WizardTasks, EntryItems);
    end else
      ProcessTask := True;

    EntryItems.Free();

    Result := ProcessComponent and ProcessTask;
  end else
    Result := True;
end;

function ShouldProcessRunEntry(const WizardComponents, WizardTasks: TStringList;
  const RunEntry: PSetupRunEntry): Boolean;
begin
  if (SetupHeader.InstallMode <> imNormal) and (roSkipIfSilent in RunEntry.Options) then
    Result := False
  else if (SetupHeader.InstallMode = imNormal) and (roSkipIfNotSilent in RunEntry.Options) then
    Result := False
  else
    Result := ShouldProcessEntry(WizardComponents, WizardTasks, RunEntry.Components, RunEntry.Tasks);
end;

function ShouldProcessIconEntry(const WizardComponents, WizardTasks: TStringList;
  const WizardNoIcons: Boolean; const IconEntry: PSetupIconEntry): Boolean;
begin
  if WizardNoIcons and (IconEntry.Tasks = '') then
    Result := False
  else
    Result := ShouldProcessEntry(WizardComponents, WizardTasks, IconEntry.Components, IconEntry.Tasks);
end;

procedure LoadInf(const FileName: String);
var
  Section: String;
begin
  Section := 'Setup';
  InitDir := GetIniString(Section, 'Dir', InitDir, FileName);
  InitProgramGroup := GetIniString(Section, 'Group', InitProgramGroup, FileName);
  InitNoIcons := GetIniBool(Section, 'NoIcons', InitNoIcons, FileName);
  InitSilent := GetIniBool(Section, 'Silent', InitSilent, FileName);
  InitVerySilent := GetIniBool(Section, 'VerySilent', InitVerySilent, FileName);
  InitNoRestart := GetIniBool(Section, 'NoRestart', InitNoRestart, FileName);
  InitComponents.CommaText := GetIniString(Section, 'Components', InitComponents.CommaText, FileName);
  InitSaveInf := GetIniString(Section, 'SaveInf', InitSaveInf, FileName);
end;

procedure SaveInf(const FileName: String);
var
  Section: String;
begin
  Section := 'Setup';
  SetIniString(Section, 'Dir', WizardDirValue, FileName);
  SetIniString(Section, 'Group', WizardGroupValue, FileName);
  SetIniBool(Section, 'NoIcons', WizardNoIcons, FileName);
  SetIniString(Section, 'Components', WizardComponents.CommaText, FileName);
end;

function ChangeIndividualConst (const Cnst: String; const CustomConsts: array of String;
  var IsPath: Boolean): String;
{ Cnst must be the name of a single constant, without the braces.
  For example: app
  IsPath is set to True if the result is a path which needs special trailing-
  backslash handling. }

  function ExpandEnvConst(C: String): String;
  var
    I: Integer;
    VarName, Default: String;
  begin
    Delete(C, 1, 1);
    I := ConstPos('|', C);  { check for 'default' value }
    if I = 0 then
      I := Length(C)+1;
    VarName := Copy(C, 1, I-1);
    Default := Copy(C, I+1, Maxint);
    Result := '';
    if ConvertConstPercentStr(VarName) and ConvertConstPercentStr(Default) then begin
      Result := GetEnv(ChangeDirConstEx(VarName, CustomConsts));
      if Result = '' then
        Result := ChangeDirConstEx(Default, CustomConsts);
    end;
  end;

  function ExpandRegConst (C: String): String;
  { Expands a registry-value constant in the form:
    reg:HKxx\SubkeyName,ValueName|DefaultValue }
  type
    TKeyNameConst = packed record
      KeyName: String;
      KeyConst: HKEY;
    end;
  const
    KeyNameConsts: array[0..4] of TKeyNameConst = (
      (KeyName: 'HKCR'; KeyConst: HKEY_CLASSES_ROOT),
      (KeyName: 'HKCU'; KeyConst: HKEY_CURRENT_USER),
      (KeyName: 'HKLM'; KeyConst: HKEY_LOCAL_MACHINE),
      (KeyName: 'HKU';  KeyConst: HKEY_USERS),
      (KeyName: 'HKCC'; KeyConst: HKEY_CURRENT_CONFIG));
  var
    Z, Subkey, Value, Default: String;
    I, J: Integer;
    RootKey: HKEY;
    K: HKEY;
  begin
    Delete (C, 1, 4);  { skip past 'reg:' }
    I := ConstPos('\', C);
    if I <> 0 then begin
      Z := Copy(C, 1, I-1);
      if Z <> '' then begin
        RootKey := 0;
        for J := Low(KeyNameConsts) to High(KeyNameConsts) do
          if CompareText(KeyNameConsts[J].KeyName, Z) = 0 then begin
            RootKey := KeyNameConsts[J].KeyConst;
            Break;
          end;
        if RootKey <> 0 then begin
          Z := Copy(C, I+1, Maxint);
          I := ConstPos('|', Z);  { check for a 'default' data }
          if I = 0 then
            I := Length(Z)+1;
          Default := Copy(Z, I+1, Maxint);
          SetLength (Z, I-1);
          I := ConstPos(',', Z);  { comma separates subkey and value }
          if I <> 0 then begin
            Subkey := Copy(Z, 1, I-1);
            Value := Copy(Z, I+1, Maxint);
            if ConvertConstPercentStr(Subkey) and ConvertConstPercentStr(Value) and
               ConvertConstPercentStr(Default) then begin
              Result := ChangeDirConstEx(Default, CustomConsts);
              if RegOpenKeyEx(RootKey, PChar(ChangeDirConstEx(Subkey, CustomConsts)),
                 0, KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
                RegQueryStringValue (K, PChar(ChangeDirConstEx(Value, CustomConsts)),
                  Result);
                RegCloseKey (K);
              end;
              Exit;
            end;
          end;
        end;
      end;
    end;
    { it will only reach here if there was a parsing error }
    InternalError('Failed to parse "reg" constant');
  end;

  function ExpandIniConst (C: String): String;
  { Expands an INI-value constant in the form:
    filename,section,key|defaultvalue }
  var
    Z, Filename, Section, Key, Default: String;
    I: Integer;
  begin
    Delete (C, 1, 4);  { skip past 'ini:' }
    I := ConstPos(',', C);
    if I <> 0 then begin
      Z := Copy(C, 1, I-1);
      if Z <> '' then begin
        Filename := Z;
        Z := Copy(C, I+1, Maxint);
        I := ConstPos('|', Z);  { check for a 'default' data }
        if I = 0 then
          I := Length(Z)+1;
        Default := Copy(Z, I+1, Maxint);
        SetLength (Z, I-1);
        I := ConstPos(',', Z);  { comma separates section and key }
        if I <> 0 then begin
          Section := Copy(Z, 1, I-1);
          Key := Copy(Z, I+1, Maxint);
          if ConvertConstPercentStr(Filename) and ConvertConstPercentStr(Section) and ConvertConstPercentStr(Key) and
             ConvertConstPercentStr(Default) then begin
            Filename := ChangeDirConstEx(Filename, CustomConsts);
            Section := ChangeDirConstEx(Section, CustomConsts);
            Key := ChangeDirConstEx(Key, CustomConsts);
            Default := ChangeDirConstEx(Default, CustomConsts);
            Result := GetIniString(Section, Key, Default, Filename);
            Exit;
          end;
        end;
      end;
    end;
    { it will only reach here if there was a parsing error }
    InternalError('Failed to parse "ini" constant');
  end;

  function ExpandParamConst (C: String): String;
  { Expands an commandline-parameter-value constant in the form:
    parametername|defaultvalue }

    function GetParamString(const StartParam: Integer; const Param, Default: String): String;
    var
      I, PCount: Integer;
      Z: String;
    begin
      PCount := NewParamCount();
      for i := StartParam to PCount do begin
        Z := NewParamStr(i);
        if StrLIComp(PChar(Z), PChar('/'+Param+'='), Length(Param)+2) = 0 then begin
          Delete(Z, 1, Length(Param)+2);
          Result := Z;
          Exit;
        end;
      end;

      Result := Default;
    end;

  var
    Z, Param, Default: String;
    I: Integer;
  begin
    Delete (C, 1, 6);  { skip past 'param:' }
    Z := C;
    I := ConstPos('|', Z);  { check for a 'default' data }
    if I = 0 then
      I := Length(Z)+1;
    Default := Copy(Z, I+1, Maxint);
    SetLength (Z, I-1);
    Param := Z;
    if ConvertConstPercentStr(Param) and ConvertConstPercentStr(Default) then begin
      Param := ChangeDirConstEx(Param, CustomConsts);
      Default := ChangeDirConstEx(Default, CustomConsts);
      Result := GetParamString(StartParam, Param, Default);
      Exit;
    end;
    { it will only reach here if there was a parsing error }
    InternalError('Failed to parse "param" constant');
  end;

const
  FolderConsts: array[Boolean, TShellFolderID] of String =
    (('userdesktop', 'userstartmenu', 'userprograms', 'userstartup',
      'sendto', 'fonts', 'userappdata', 'userdocs', 'usertemplates',
      'userfavorites', 'localappdata'),
     ('commondesktop', 'commonstartmenu', 'commonprograms', 'commonstartup',
      'sendto', 'fonts', 'commonappdata', 'commondocs', 'commontemplates',
      'commonfavorites', 'localappdata'));
var
  Z: String;
  B: Boolean;
  SF: TShellFolderID;
  K: Integer;
begin
  { Non-path constants }
  IsPath := False;
  if Cnst = '\' then Result := '\'
  else if Cnst = 'hwnd' then Result := IntToStr(MainForm.Handle)
  else if Cnst = 'computername' then Result := GetComputerNameString
  else if Cnst = 'username' then Result := GetUserNameString
  else if Cnst = 'groupname' then Result := WizardGroupValue
  else if Cnst = 'wizardhwnd' then begin
    if Assigned(WizardForm) then
      Result := IntToStr(WizardForm.Handle)
    else
      Result := '0';
  end else begin
    { Path constants }
    IsPath := True;
    if Cnst[1] = '%' then Result := ExpandEnvConst(Cnst)
    else if Copy(Cnst, 1, 4) = 'reg:' then Result := ExpandRegConst(Cnst)
    else if Copy(Cnst, 1, 4) = 'ini:' then Result := ExpandIniConst(Cnst)
    else if Copy(Cnst, 1, 6) = 'param:' then Result := ExpandParamConst(Cnst)
    else if Cnst = 'src' then Result := SourceDir
    else if Cnst = 'srcexe' then Result := SetupLdrOriginalFilename
    else if Cnst = 'tmp' then Result := TempInstallDir
    else if Cnst = 'app' then Result := WizardDirValue
    else if Cnst = 'win' then Result := WinDir
    else if Cnst = 'sys' then Result := WinSystemDir
    else if Cnst = 'sd' then Result := SystemDrive
    else if Cnst = 'pf' then Result := ProgramFilesDir
    else if Cnst = 'cf' then Result := CommonFilesDir
    else if Cnst = 'dao' then Result := DAODir
    else if Cnst = 'cmd' then Result := CmdFilename
    else if Cnst = 'sysuserinfoname' then Result := SysUserInfoName
    else if Cnst = 'sysuserinfoorg' then Result := SysUserInfoOrg
    else if Cnst = 'userinfoname' then Result := WizardUserInfoName
    else if Cnst = 'userinfoorg' then Result := WizardUserInfoOrg
    else if Cnst = 'uninstallexe' then Result := UninstallExeFilename
    else if Cnst = 'group' then begin
      if WizardGroupValue <> '' then begin
        Z := GetShellFolder(not(shAlwaysUsePersonalGroup in SetupHeader.Options),
          sfPrograms, ConstReadOnly);
        if Z = '' then
          InternalError('Failed to expand "group" constant');
        Result := AddBackslash(Z) + WizardGroupValue;
      end
      else
        Result := '';
    end
    else begin
      { Shell folder constants }
      for B := False to True do
        for SF := Low(SF) to High(SF) do
          if Cnst = FolderConsts[B, SF] then begin
            Z := GetShellFolder(B, SF, ConstReadOnly);
            if Z = '' then begin
              if SF <> sfFonts then
                InternalError(Format('Failed to expand shell folder constant "%s"', [Cnst]))
              else
                Result := FontDir;
            end
            else
              Result := Z;
            Exit;
          end;
      { Custom constants (non-path) }
      IsPath := False;
      if Cnst <> '' then begin
        K := 0;
        while K < High(CustomConsts) do begin
          if Cnst = CustomConsts[K] then begin
            Result := CustomConsts[K+1];
            Exit;
          end;
          Inc (K, 2);
        end;
      end;
      { Unknown constant }
      InternalError(Format('Unknown constant "%s"', [Cnst]));
    end;
  end;
end;

function ChangeDirConst (const S: String): String;
begin
  Result := ChangeDirConstEx(S, ['']);
end;

function ChangeDirConstEx (const S: String; const CustomConsts: array of String): String;
var
  I, Start: Integer;
  Cnst, ReplaceWith: String;
  IsPath: Boolean;
begin
  Result := S;
  I := 1;
  while I <= Length(Result) do begin
    if Result[I] = '{' then begin
      if (I < Length(Result)) and (Result[I+1] = '{') then begin
        { Change '{{' to '{' if not in an embedded constant }
        Inc (I);
        Delete (Result, I, 1);
      end
      else begin
        Start := I;
        { Find the closing brace, skipping over any embedded constants }
        I := SkipPastConst(Result, I);
        if I = 0 then  { unclosed constant? }
          InternalError('Unclosed constant');
        Dec (I);  { 'I' now points to the closing brace }

        { Now translate the constant }
        Cnst := Copy(Result, Start+1, I-(Start+1));
        ReplaceWith := ChangeIndividualConst(Cnst, CustomConsts, IsPath);
        Delete (Result, Start, (I+1)-Start);
        Insert (ReplaceWith, Result, Start);
        I := Start + Length(ReplaceWith);
        if IsPath and (ReplaceWith <> '') and
           (AnsiLastChar(ReplaceWith)^ = '\') and
           (I <= Length(Result)) and (Result[I] = '\') then
          Delete (Result, I, 1);
      end;
    end
    else begin
      if Result[I] in ConstLeadBytes^ then
        Inc (I);
      Inc (I);
    end;
  end;
end;

function GetShellFolder (Common: Boolean; const ID: TShellFolderID;
  ReadOnly: Boolean): String;

  procedure GetFolder (const Common: Boolean);
  const
    CSIDL_COMMON_STARTMENU = $0016;
    CSIDL_COMMON_PROGRAMS = $0017;
    CSIDL_COMMON_STARTUP = $0018;
    CSIDL_COMMON_DESKTOPDIRECTORY = $0019;
    CSIDL_APPDATA = $001A;
    CSIDL_LOCAL_APPDATA = $001C;
    CSIDL_COMMON_FAVORITES = $001F;
    CSIDL_COMMON_APPDATA = $0023;
    CSIDL_COMMON_TEMPLATES = $002D;
    CSIDL_COMMON_DOCUMENTS = $002E;
    CSIDL_FLAG_CREATE = $8000;
    SHGFP_TYPE_CURRENT = 0;
    FolderIDs: array[Boolean, TShellFolderID] of Integer = (
      { User }
      (CSIDL_DESKTOPDIRECTORY, CSIDL_STARTMENU, CSIDL_PROGRAMS, CSIDL_STARTUP,
       CSIDL_SENDTO, CSIDL_FONTS, CSIDL_APPDATA, CSIDL_PERSONAL,
       CSIDL_TEMPLATES, CSIDL_FAVORITES, CSIDL_LOCAL_APPDATA),
      { Common }
      (CSIDL_COMMON_DESKTOPDIRECTORY, CSIDL_COMMON_STARTMENU, CSIDL_COMMON_PROGRAMS, CSIDL_COMMON_STARTUP,
       CSIDL_SENDTO, CSIDL_FONTS, CSIDL_COMMON_APPDATA, CSIDL_COMMON_DOCUMENTS,
       CSIDL_COMMON_TEMPLATES, CSIDL_COMMON_FAVORITES, CSIDL_LOCAL_APPDATA));
  var
    Folder: Integer;
    Z: String;
    Buf: array[0..MAX_PATH-1] of Char;
  begin
    if not ShellFoldersRead[Common, ID] then begin
      Folder := FolderIDs[Common, ID];
      if not ReadOnly then begin
        { CSIDL_FLAG_CREATE is needed or else the function fails if the
          specified CSIDL is valid but doesn't currently exist. }
        Folder := Folder or CSIDL_FLAG_CREATE;
      end;
      if SUCCEEDED(SHGetFolderPathFunc(0, Folder, 0, SHGFP_TYPE_CURRENT, Buf)) then
        Z := RemoveBackslashUnlessRoot(Buf);
      ShellFolders[Common, ID] := Z;
      if not ReadOnly or (Z <> '') then
        ShellFoldersRead[Common, ID] := True;
    end;
    Result := ShellFolders[Common, ID];
  end;

begin
  Result := '';
  { If user isn't an administrator, always use User constants }
  if not IsAdmin then
    Common := False;
  { Don't try to use certain Common constants on Windows 9x, because the shell
    doesn't know what to do with them }
  if Common and not IsNT and
     (ID in [sfDesktop, sfStartMenu, sfPrograms, sfStartup]) then
    Common := False;
  GetFolder (Common);
  if (Result = '') and Common then
    { If it failed to get the path of a Common CSIDL, try getting the
      User version of the CSIDL instead. (Many of the Common CSIDLS are
      unsupported by Win9x.) }
    GetFolder (False);
end;

function InstallOnThisVersion (const MinVersion: TSetupVersionData;
  const OnlyBelowVersion: TSetupVersionData): TInstallOnThisVersionResult;
var
  Ver, Ver2, MinVer, OnlyBelowVer: Cardinal;
begin
  Ver := WindowsVersion;
  if IsNT then begin
    MinVer := MinVersion.NTVersion;
    OnlyBelowVer := OnlyBelowVersion.NTVersion;
  end
  else begin
    MinVer := MinVersion.WinVersion;
    OnlyBelowVer := OnlyBelowVersion.WinVersion;
  end;
  Result := irInstall;
  if MinVer = 0 then
    Result := irNotOnThisPlatform
  else begin
    if (Ver < MinVer) or
       (IsNT and (LongRec(Ver).Hi = LongRec(MinVer).Hi) and
        (NTServicePackLevel < MinVersion.NTServicePack)) then
      Result := irVerTooLow
    else begin
      if OnlyBelowVer <> 0 then begin
        Ver2 := Ver;
        { A build number of 0 on OnlyBelowVersion means 'match any build' }
        if LongRec(OnlyBelowVer).Lo = 0 then
          Ver2 := Ver2 and $FFFF0000;  { set build number to zero on Ver2 also }
        if not IsNT then begin
          if Ver2 >= OnlyBelowVer then
            Result := irVerTooHigh;
        end
        else begin
          if (Ver2 > OnlyBelowVer) or
             ((LongRec(Ver).Hi = LongRec(OnlyBelowVer).Hi) and
              (NTServicePackLevel >= OnlyBelowVersion.NTServicePack)) then
            Result := irVerTooHigh;
        end;
      end;
    end;
  end;
end;

function GetSizeOfFiles (const Filename: String): Longint;
{ Returns sizes of all files matching Filename, or -1 if an error occured }
var
  H: THandle;
  FindData: TWin32FindData;
begin
  H := FindFirstFile(PChar(Filename), FindData);
  if H <> INVALID_HANDLE_VALUE then begin
    Result := 0;
    repeat
      if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0 then begin
        if FindData.nFileSizeHigh <> 0 then begin  { file size > 4GB ! }
          Result := High(Longint);  { just return maximum size of a Longint (2GB) }
          Break;
        end
        else begin
          { If > 2GB (which is stored as a negative number on a signed 32-bit int),
            or if adding size to Result will cause it to wrap around (exceed 2GB),
            return 2GB }
          if (Longint(FindData.nFileSizeLow) < 0) or
             (Result + Longint(FindData.nFileSizeLow) < Result) then begin
            Result := High(Longint);
            Break;
          end;
          Inc (Result, FindData.nFileSizeLow);
        end;
      end;
    until not FindNextFile(H, FindData);
    Windows.FindClose (H);
  end
  else
    Result := -1;
end;

function GetSizeOfComponent(const ComponentName: String; const ExtraDiskSpaceRequired: Longint): Longint;
var
  FileComponents: TStringList;
  ComponentSize: LongInt;
  I: Integer;
begin
  ComponentSize := ExtraDiskSpaceRequired;
  FileComponents := TStringList.Create();

  for I := 0 to Entries[seFile].Count-1 do begin
    with PSetupFileEntry(Entries[seFile][I])^ do begin
      if (Tasks = '') then begin { don't count tasks }
        FileComponents.CommaText := Components;
        if FileComponents.IndexOf(ComponentName) <> -1 then begin
          if LocationEntry <> -1 then
            Inc(ComponentSize, PSetupFileLocationEntry(Entries[seFileLocation][LocationEntry])^.OriginalSize)
          else
            Inc(ComponentSize, ExternalSize)
        end;
      end;
    end;
  end;

  FileComponents.Free();
  Result := ComponentSize;
end;

function GetSizeOfType(const TypeName: String; const IsCustom: Boolean): LongInt;
var
  ComponentTypes: TStringList;
  TypeSize: LongInt;
  I: Integer;
begin
  TypeSize := 0;
  ComponentTypes := TStringList.Create();

  for I := 0 to Entries[seComponent].Count-1 do begin
    with PSetupComponentEntry(Entries[seComponent][I])^ do begin
      ComponentTypes.CommaText := Types;
      { For custom types, only count fixed components. Otherwise count all. }
      if IsCustom then begin
        if (coFixed in Options) and (ComponentTypes.IndexOf(TypeName) <> -1) then
          Inc(TypeSize, Size);
      end else begin
        if ComponentTypes.IndexOf(TypeName) <> -1 then
          Inc(TypeSize, Size);
      end;
    end;
  end;

  ComponentTypes.Free();
  Result := TypeSize;
end;

procedure DebugNotifyEntry(EntryType: TEntryType; Number: Integer);
var
  Kind: TDebugEntryKind;
begin
  if not Debugging then Exit;
  case EntryType of
    seDir: Kind := deDir;
    seFile: Kind := deFile;
    seIcon: Kind := deIcon;
    seIni: Kind := deIni;
    seRegistry: Kind := deRegistry;
    seInstallDelete: Kind := deInstallDelete;
    seUninstallDelete: Kind := deUninstallDelete;
    seRun: Kind := deRun;
    seUninstallRun: Kind := deUninstallRun;
  else
    Exit;
  end;
  DebugNotify(Kind, Integer(OriginalEntryIndexes[EntryType][Number]));
end;

procedure InternalError (const Id: String);
begin
  raise Exception.Create(FmtSetupMessage1(msgErrorInternal2, Id));
end;

procedure InitializeSetup;
{ Initializes various vars used by the setup. This is called in the project
  source. }

  function VerToStr (Ver: Cardinal; ServicePack: Word): String;
  var
    Digits: Integer;
  begin
    with TSetupVersionDataVersion(Ver) do begin
      Digits := 2;
      if Minor mod 10 = 0 then begin
        Dec (Digits);
        Minor := Minor div 10;
      end;
      FmtStr (Result, '%d.%.*d', [Major, Digits, Minor]);
      if Build <> 0 then
        Result := Result + Format('.%d', [Build]);
      if ServicePack <> 0 then begin
        Result := Result + ' Service Pack ' + IntToStr(Hi(ServicePack));
        if Lo(ServicePack) <> 0 then
          Result := Result + Format('.%d', [Lo(ServicePack)]);
      end;
    end;
  end;

  procedure AbortInit (const Msg: TSetupMessageID);
  begin
    MsgBox (SetupMessages[Msg], '', mbCriticalError, MB_OK);
    Abort;
  end;

  procedure AbortInitFmt1 (const Msg: TSetupMessageID; const Arg1: String);
  begin
    MsgBox (FmtSetupMessage(Msg, [Arg1]), '', mbCriticalError, MB_OK);
    Abort;
  end;

  procedure AbortInitVerError (const Msg: TSetupMessageID; const Platform: String;
    const Ver: Cardinal; const ServicePack: Word);
  begin
    MsgBox (FmtSetupMessage(Msg, [Platform, VerToStr(Ver, ServicePack)]), '',
      mbCriticalError, MB_OK);
    Abort;
  end;

  function CreateTempDir: String;
  var
    ErrorCode: DWORD;
  begin
    Result := GenerateUniqueName(GetTempDir, '.tmp');
    if not CreateDirectory(PChar(Result), nil) then begin
      ErrorCode := GetLastError;
      raise Exception.Create(FmtSetupMessage(msgLastErrorMessage,
        [FmtSetupMessage1(msgErrorCreatingDir, Result), IntToStr(ErrorCode),
         SysErrorMessage(ErrorCode)]));
    end;
  end;

  procedure SaveStreamToTempFile (const Strm: TCustomMemoryStream;
    const Filename: String);
  var
    ErrorCode: DWORD;
  begin
    try
      Strm.SaveToFile (Filename);
    except
      { Display more useful error message than 'Stream write error' etc. }
      on EStreamError do begin
        ErrorCode := GetLastError;
        raise Exception.Create(FmtSetupMessage(msgLastErrorMessage,
          [SetupMessages[msgLdrCannotCreateTemp], IntToStr(ErrorCode),
           SysErrorMessage(ErrorCode)]));
      end;
    end;
  end;

  procedure LoadSHFolderDLL;
  var
    Filename: String;
    ResStrm: TResourceStream;
    ExistingFileVersion, NewFileVersion: TFileVersionNumbers;
  const
    shfolder = 'shfolder.dll';
  begin
    Filename := AddBackslash(TempInstallDir) + '_shfoldr.dll';
    {$R _shfoldr.res}  { Link in the .res file containing the DLL image }
    ResStrm := TResourceStream.Create(HInstance, 'SHFOLDERDLL', RT_RCDATA);
    try
      SaveStreamToTempFile (ResStrm, Filename);
    finally
      ResStrm.Free;
    end;
    if not GetVersionNumbers(Filename, NewFileVersion) then
      InternalError('Failed to get version numbers of _shfoldr.dll');
    { Does the system already have the same version or a newer version of
      shfolder.dll? If so, use it instead of the one we just extracted. }
    if GetVersionNumbers(shfolder, ExistingFileVersion) and
       (((ExistingFileVersion.MS > NewFileVersion.MS) or
         ((ExistingFileVersion.MS = NewFileVersion.MS) and
          (ExistingFileVersion.LS > NewFileVersion.LS)))) or
        ((ExistingFileVersion.MS = NewFileVersion.MS) and
         (ExistingFileVersion.LS = NewFileVersion.LS)) then
      Filename := shfolder;
    { Ensure shell32.dll is pre-loaded so it isn't loaded/freed for each
      individual SHGetFolderPath call }
    SHFolderShell32Handle := LoadLibrary(shell32);
    SHFolderDLLHandle := LoadLibrary(PChar(Filename));
    if SHFolderDLLHandle = 0 then
      InternalError(Format('Failed to load DLL "%s"', [Filename]));
    @SHGetFolderPathFunc := GetProcAddress(SHFolderDLLHandle, 'SHGetFolderPathA');
    if @SHGetFolderPathFunc = nil then
      InternalError('Failed to get address of SHGetFolderPathA function');
  end;

  procedure ReadFileIntoStream (const Stream: TStream;
    const Data: PDeflateBlockReadData);
  type
    PBuffer = ^TBuffer;
    TBuffer = array[0..8191] of Byte;
  var
    Buf: PBuffer;
    BytesLeft, Bytes: Longint;
  begin
    New (Buf);
    try
      InflateBlockRead (Data^, BytesLeft, SizeOf(BytesLeft));
      while BytesLeft > 0 do begin
        Bytes := BytesLeft;
        if Bytes > SizeOf(Buf^) then Bytes := SizeOf(Buf^);
        InflateBlockRead (Data^, Buf^, Bytes);
        Stream.WriteBuffer (Buf^, Bytes);
        Dec (BytesLeft, Bytes);
      end;
    finally
      Dispose (Buf);
    end;
  end;

  procedure ReadWizardImage (var WizardImage: TBitmap; const Data: PDeflateBlockReadData);
  var
    MemStream: TMemoryStream;
  begin
    MemStream := TMemoryStream.Create;
    try
      ReadFileIntoStream (MemStream, Data);
      MemStream.Seek (0, soFromBeginning);
      WizardImage := TBitmap.Create;
      WizardImage.LoadFromStream (MemStream);
    finally
      MemStream.Free;
    end;
  end;

  procedure LoadBunzipDLL;
  var
    Filename: String;
  begin
    Filename := AddBackslash(TempInstallDir) + '_isbunzp.dll';
    SaveStreamToTempFile (BunzipDLL, Filename);
    BunzipDLL.Free;
    BunzipDLL := nil;
    BunzipDLLHandle := LoadLibrary(PChar(Filename));
    if BunzipDLLHandle = 0 then
      InternalError(Format('Failed to load DLL "%s"', [Filename]));
    if not BZInitDecompressFunctions(BunzipDLLHandle) then
      InternalError('BZInitDecompressFunctions failed');
  end;

  procedure ReadSysUserInfo;
  const
    Paths: array[Boolean] of PChar = (NEWREGSTR_PATH_SETUP,
      'SOFTWARE\Microsoft\Windows NT\CurrentVersion');
  var
    K: HKEY;
  begin
    if RegOpenKeyEx(HKEY_LOCAL_MACHINE, Paths[IsNT], 0, KEY_QUERY_VALUE,
       K) = ERROR_SUCCESS then begin
      RegQueryStringValue (K, 'RegisteredOwner', SysUserInfoName);
      RegQueryStringValue (K, 'RegisteredOrganization', SysUserInfoOrg);
      RegCloseKey (K);
    end;
  end;

var
  PCount: Integer;

  SetupFilename: String;
  SetupFile: File;
  TestID: TSetupID;
  Data: PDeflateBlockReadData;
  BeganInflating: Boolean;

  I: Integer;

  Name: String;
  F: File;

  CurFileSize: Longint;
  OldErrorMode: UINT;

  procedure ReadEntries (const EntryType: TEntryType; const Count: Integer;
    const Size: Integer; const MinVersionOfs, OnlyBelowVersionOfs: Integer);
  var
    I: Integer;
    P: Pointer;
  begin
    if Debugging then begin
      OriginalEntryIndexes[EntryType] := TList.Create;
      OriginalEntryIndexes[EntryType].Capacity := Count;
    end;
    for I := 0 to Count-1 do begin
      P := AllocMem(Size);
      SEInflateBlockRead (Data^, P^, Size, EntryStrings[EntryType]);
      if InstallOnThisVersion(TSetupVersionData((@PByteArray(P)[MinVersionOfs])^),
         TSetupVersionData((@PByteArray(P)[OnlyBelowVersionOfs])^)) = irInstall then begin
        Entries[EntryType].Add(P);
        if Debugging then
          OriginalEntryIndexes[EntryType].Add(Pointer(I));
      end
      else
        SEFreeRec (P, EntryStrings[EntryType]);
    end;
  end;

var
  NewEntry: Pointer;
  MinimumTypeSpace: LongInt;
begin
  { Based on SetupLdr or not?
    Parameters for launching SetupLdr-based installation are:
    /SL3 <handle to SetupLdrWnd> <original exe filename> <setup msg data offset>
         <setup 0 data offset> <setup 1 data offset>
  }
  PCount := NewParamCount;
  if (PCount >= 6) and (CompareText(NewParamStr(1), '/SL3') = 0) then begin
    StartParam := 7;
    SetupLdrMode := True;
    SetupLdrWnd := StrToInt(NewParamStr(2));
    SetupLdrOriginalFilename := NewParamStr(3);
    SetupLdrOffsetMsg := StrToInt(NewParamStr(4));
    LoadCompressedSetupMessages (SetupLdrOriginalFilename, SetupLdrOffsetMsg,
      fmOpenRead or fmShareDenyWrite);
    SourceDir := ExtractFilePath(SetupLdrOriginalFilename);
    SetupLdrOffset0 := StrToInt(NewParamStr(5));
    SetupLdrOffset1 := StrToInt(NewParamStr(6));
  end
  else begin
    StartParam := 1;
    LoadCompressedSetupMessages (ExtractFilePath(ParamStr(0)) + 'SETUP.MSG', 0,
      fmOpenRead or fmShareDenyWrite);
    SetupLdrOriginalFilename := NewParamStr(0);
    SourceDir := ExtractFilePath(SetupLdrOriginalFilename);
  end;
  SourceDir := RemoveBackslashUnlessRoot(SourceDir);

  SetMessageBoxCaption (mbInformation, PChar(SetupMessages[msgInformationTitle]));
  SetMessageBoxCaption (mbConfirmation, PChar(SetupMessages[msgConfirmTitle]));
  SetMessageBoxCaption (mbError, PChar(SetupMessages[msgErrorTitle]));
  SetMessageBoxCaption (mbCriticalError, PChar(SetupMessages[msgErrorTitle]));

  if Win32Platform = VER_PLATFORM_WIN32s then
    AbortInitFmt1 (msgNotOnThisPlatform, 'Win32s');

  Application.Title := SetupMessages[msgSetupAppTitle];

  IsAdmin := IsAdminLoggedOn;
  IsPowerUser := IsAdmin or IsPowerUserLoggedOn;
  Randomize;

  for I := StartParam to PCount do begin
    Name := NewParamStr(I);
    if CompareText(Name, '/Silent') = 0 then
      InitSilent := True
    else
    if CompareText(Name, '/VerySilent') = 0 then
      InitVerySilent := True
    else
    if CompareText(Name, '/NoRestart') = 0 then
      InitNoRestart := True
    else
    if CompareText(Name, '/NoIcons') = 0 then
      InitNoIcons := True
    else
    if CompareText(Name, '/NoCancel') = 0 then
      InitNoCancel := True
    else
    if CompareText(Copy(Name, 1, 12), '/Components=') = 0 then
      InitComponents.CommaText := Copy(Name, 13, Maxint)
    else
    if CompareText(Copy(Name, 1, 9), '/LoadInf=') = 0 then
      InitLoadInf := Copy(Name, 10, Maxint)
    else
    if CompareText(Copy(Name, 1, 9), '/SaveInf=') = 0 then
      InitSaveInf := Copy(Name, 10, Maxint)
    else
    if CompareText(Copy(Name, 1, 5), '/DIR=') = 0 then
      InitDir := Copy(Name, 6, Maxint)
    else
    if CompareText(Copy(Name, 1, 7), '/GROUP=') = 0 then
      InitProgramGroup := Copy(Name, 8, Maxint)
    else
    if CompareText(Name, '/DETACHEDMSG') = 0 then  { for debugging }
      DetachedUninstMsgFile := True
    else
    if CompareText(Copy(Name, 1, 10), '/DEBUGWND=') = 0 then
      SetDebugWnd(StrToInt(Copy(Name, 11, Maxint)));
  end;

  if InitLoadInf <> '' then
    LoadInf(InitLoadInf);

  { Read SETUP.0, or from EXE }
  if not SetupLdrMode then begin
    SetupFilename := AddBackslash(SourceDir) + 'SETUP.0'; {don't localize}
    if not NewFileExists(SetupFilename) then
      AbortInitFmt1 (msgSetupFileMissing, 'SETUP.0');
  end
  else
    SetupFilename := SetupLdrOriginalFilename;
  AssignFile (SetupFile, SetupFilename);
  FileMode := fmOpenRead or fmShareDenyWrite;  Reset (SetupFile, 1);
  try
    Seek (SetupFile, SetupLdrOffset0);
    BlockRead (SetupFile, TestID, SizeOf(TestID));
    if TestID <> SetupID then
      AbortInit (msgSetupFileCorruptOrWrongVer);
    try
      BeganInflating := False;
      New (Data);
      try
        InflateBlockReadBegin (SetupFile, Data^);
        BeganInflating := True;
        { Header }
        SEInflateBlockRead (Data^, SetupHeader, SizeOf(SetupHeader),
          SetupHeaderStrings);
        SEInflateBlockRead (Data^, LangOptions, SizeOf(LangOptions),
          SetupLangOptionsStrings);
        case InstallOnThisVersion(SetupHeader.MinVersion, SetupHeader.OnlyBelowVersion) of
          irInstall: ;
          irNotOnThisPlatform:
            if IsNT then
              AbortInitFmt1 (msgNotOnThisPlatform, 'Windows NT')
            else
              AbortInitFmt1 (msgOnlyOnThisPlatform, 'Windows NT');
          irVerTooLow:
            if IsNT then
              AbortInitVerError (msgWinVersionTooLowError, 'Windows NT',
                SetupHeader.MinVersion.NTVersion,
                SetupHeader.MinVersion.NTServicePack)
            else
              AbortInitVerError (msgWinVersionTooLowError, 'Windows',
                SetupHeader.MinVersion.WinVersion, 0);
          irVerTooHigh:
            if IsNT then
              AbortInitVerError (msgWinVersionTooHighError, 'Windows NT',
                SetupHeader.OnlyBelowVersion.NTVersion,
                SetupHeader.OnlyBelowVersion.NTServicePack)
            else
              AbortInitVerError (msgWinVersionTooHighError, 'Windows',
                SetupHeader.OnlyBelowVersion.WinVersion, 0);
        end;
        NeedPassword := shPassword in SetupHeader.Options;
        case SetupHeader.PrivilegesRequired of
          prPowerUser:
            if not IsPowerUser then AbortInit (msgPowerUserPrivilegesRequired);
          prAdmin:
            if not IsAdmin then AbortInit (msgAdminPrivilegesRequired);
        end;
        NeedsRestart := shAlwaysRestart in SetupHeader.Options;
        { Wizard image }
        ReadWizardImage (WizardImage, Data);
        ReadWizardImage (WizardSmallImage, Data);
        { bunzip DLL }
        if shBzipUsed in SetupHeader.Options then begin
          BunzipDLL := TMemoryStream.Create;
          ReadFileIntoStream (BunzipDLL, Data);
        end;
        { Type entries }
        ReadEntries (seType, SetupHeader.NumTypeEntries, SizeOf(TSetupTypeEntry),
          Integer(@PSetupTypeEntry(nil).MinVersion),
          Integer(@PSetupTypeEntry(nil).OnlyBelowVersion));
        { Component entries }
        ReadEntries (seComponent, SetupHeader.NumComponentEntries, SizeOf(TSetupComponentEntry),
          Integer(@PSetupComponentEntry(nil).MinVersion),
          Integer(@PSetupComponentEntry(nil).OnlyBelowVersion));
        { Task entries }
        ReadEntries (seTask, SetupHeader.NumTaskEntries, SizeOf(TSetupTaskEntry),
          Integer(@PSetupTaskEntry(nil).MinVersion),
          Integer(@PSetupTaskEntry(nil).OnlyBelowVersion));
        { Dir entries }
        ReadEntries (seDir, SetupHeader.NumDirEntries, SizeOf(TSetupDirEntry),
          Integer(@PSetupDirEntry(nil).MinVersion),
          Integer(@PSetupDirEntry(nil).OnlyBelowVersion));
        { File entries }
        ReadEntries (seFile, SetupHeader.NumFileEntries, SizeOf(TSetupFileEntry),
          Integer(@PSetupFileEntry(nil).MinVersion),
          Integer(@PSetupFileEntry(nil).OnlyBelowVersion));
        { Icon entries }
        ReadEntries (seIcon, SetupHeader.NumIconEntries, SizeOf(TSetupIconEntry),
          Integer(@PSetupIconEntry(nil).MinVersion),
          Integer(@PSetupIconEntry(nil).OnlyBelowVersion));
        { INI entries }
        ReadEntries (seIni, SetupHeader.NumIniEntries, SizeOf(TSetupIniEntry),
          Integer(@PSetupIniEntry(nil).MinVersion),
          Integer(@PSetupIniEntry(nil).OnlyBelowVersion));
        { Registry entries }
        ReadEntries (seRegistry, SetupHeader.NumRegistryEntries, SizeOf(TSetupRegistryEntry),
          Integer(@PSetupRegistryEntry(nil).MinVersion),
          Integer(@PSetupRegistryEntry(nil).OnlyBelowVersion));
        { InstallDelete entries }
        ReadEntries (seInstallDelete, SetupHeader.NumInstallDeleteEntries, SizeOf(TSetupDeleteEntry),
          Integer(@PSetupDeleteEntry(nil).MinVersion),
          Integer(@PSetupDeleteEntry(nil).OnlyBelowVersion));
        { UninstallDelete entries }
        ReadEntries (seUninstallDelete, SetupHeader.NumUninstallDeleteEntries, SizeOf(TSetupDeleteEntry),
          Integer(@PSetupDeleteEntry(nil).MinVersion),
          Integer(@PSetupDeleteEntry(nil).OnlyBelowVersion));
        { Run entries }
        ReadEntries (seRun, SetupHeader.NumRunEntries, SizeOf(TSetupRunEntry),
          Integer(@PSetupRunEntry(nil).MinVersion),
          Integer(@PSetupRunEntry(nil).OnlyBelowVersion));
        { UninstallRun entries }
        ReadEntries (seUninstallRun, SetupHeader.NumUninstallRunEntries, SizeOf(TSetupRunEntry),
          Integer(@PSetupRunEntry(nil).MinVersion),
          Integer(@PSetupRunEntry(nil).OnlyBelowVersion));

        BeganInflating := False;
        InflateBlockReadEnd (Data^);

        InflateBlockReadBegin (SetupFile, Data^);
        BeganInflating := True;

        { File location entries }
        for I := 1 to SetupHeader.NumFileLocationEntries do begin
          GetMem (NewEntry, SizeOf(TSetupFileLocationEntry));
          InflateBlockRead (Data^, NewEntry^, SizeOf(TSetupFileLocationEntry));
          Entries[seFileLocation].Add (NewEntry);
        end;
      finally
        if BeganInflating then
          InflateBlockReadEnd (Data^);
        Dispose (Data);
      end;
    except
      on EZlibDataError do
        AbortInit (msgSetupFileCorrupt);
    end;
  finally
    CloseFile (SetupFile);
  end;

  { Setup install mode }
  if not(shPassword in SetupHeader.Options) then begin
    if InitSilent then
      SetupHeader.InstallMode := imSilent
    else if InitVerySilent then
      SetupHeader.InstallMode := imVerySilent;
  end;

  if SetupHeader.InstallMode <> imNormal then begin
    if SetupHeader.InstallMode = imVerySilent then begin
      Application.ShowMainForm := False;
      {$IFNDEF Delphi3orHigher}
      { Hide the taskbar button. Not needed for Delphi 3+ since the taskbar
        button is hidden automatically when no forms have Visible=True. }
      SetWindowPos(Application.Handle, 0, 0, 0, 0, 0, SWP_NOSIZE or
        SWP_NOMOVE or SWP_NOZORDER or SWP_NOACTIVATE or SWP_HIDEWINDOW);
      {$ENDIF}
    end;
    SetupHeader.Options := SetupHeader.Options - [shWindowVisible];
  end;

  { Check if app is running }
  while CheckForMutexes(SetupHeader.AppMutex) do
    if MsgBox(FmtSetupMessage1(msgSetupAppRunningError, SetupHeader.AppName),
       SetupMessages[msgSetupAppTitle], mbError, MB_OKCANCEL) <> IDOK then
      Abort;

  { Set misc. variables }
  UninstallRegKeyBaseName := SetupHeader.AppId;
  { Uninstall registry keys can only be up to 63 characters, otherwise Win95
    ignores them. Limit to 57 since Setup will add _isXXX to the end later. }
  if Length(UninstallRegKeyBaseName) > 57 then
    { Only keep the first 48 characters, then add an tilde and the CRC
      of the original string (to make the trimmed string unique). The
      resulting string is 57 characters long. }
    FmtStr (UninstallRegKeyBaseName, '%.48s~%.8x', [UninstallRegKeyBaseName,
      GetCRC32(UninstallRegKeyBaseName[1], Length(UninstallRegKeyBaseName))]);

  HasCustomType := False;
  for I := 0 to Entries[seType].Count-1 do begin
    if toIsCustom in PSetupTypeEntry(Entries[seType][I]).Options then begin
      HasCustomType := True;
      Break;
    end;
  end;

  HasComponents := Entries[seComponent].Count <> 0;

  HasIcons := Entries[seIcon].Count <> 0;

  HasTaskIcons := False;
  for I := 0 to Entries[seIcon].Count-1 do begin
    if PSetupIconEntry(Entries[seIcon][I]).Tasks <> '' then begin
      HasTaskIcons := True;
      Break;
    end;
  end;

  HasTasks := Entries[seTask].Count <> 0;

  { Read Windows dir }
  WinDir := GetWinDir;

  { Read Windows System dir. If it is a shared version it uses Windows dir
    instead. }
  WinSystemDir := GetSystemDir;
  if Pos(WinDir, WinSystemDir) <> 1 then begin
    { If System is not a subdirectory of Windows, possibly a shared version }
    try
      { Try making a temp file to see if the dir is write protected }
      Name := GenerateUniqueName(WinSystemDir, '.tmp');
      AssignFile (F, Name);
      FileMode := fmOpenWrite or fmShareExclusive;  Rewrite (F, 1);
      CloseFile (F);
      DeleteFile (Name);
    except
      { If exception occured, system dir is probably write protected. }
      WinSystemDir := WinDir;
    end;
  end;

  { Get system drive }
  if Win32Platform = VER_PLATFORM_WIN32_NT then
    SystemDrive := GetEnv('SystemDrive')  {don't localize}
  else
    SystemDrive := '';
  if SystemDrive = '' then begin
    SystemDrive := ExtractFileDrive(WinDir);
    if SystemDrive = '' then
      { In some rare case that ExtractFileDrive failed, just default to C }
      SystemDrive := 'C:';
  end;

  { Attempt to create a TempInstallDir }
  TempInstallDir := CreateTempDir;

  { Extract "_shfoldr.dll" to TempInstallDir, and load it }
  LoadSHFolderDLL;

  { Extract "_isbunzp.dll" to TempInstallDir, and load it }
  if shBzipUsed in SetupHeader.Options then
    LoadBunzipDLL;

  { Get Program Files and Common Files dirs }
  ProgramFilesDir := GetProgramFilesPath;
  if ProgramFilesDir = '' then
    ProgramFilesDir := SystemDrive + '\Program Files';  {don't localize}
  CommonFilesDir := GetCommonFilesPath;
  if CommonFilesDir = '' then
    CommonFilesDir := AddBackslash(ProgramFilesDir) + 'Common Files';  {don't localize}

  { Get Fonts directory }
  FontDir := GetShellFolder(False, sfFonts, False);
  if FontDir = '' then
    FontDir := WinSystemDir;

  { Generate DAO directory name }
  DAODir := AddBackslash(CommonFilesDir) + 'Microsoft Shared\DAO';

  { Get path of command interpreter }
  if IsNT then
    CmdFilename := AddBackslash(GetSystemDir) + 'cmd.exe'
  else
    CmdFilename := AddBackslash(WinDir) + 'COMMAND.COM';

  { Get user info from system }
  ReadSysUserInfo;

  { Calculate minimum disk space. If there are setup types, find the smallest
    type and add the size of all files that don't belong to any component. Otherwise
    calculate minimum disk space by adding all of the file's sizes. Also for each
    "external" file, check the file size now, and store it the ExternalSize field
    of the TSetupFileEntry record. }

  OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);  { Prevent "Network Error" boxes }
  try
    MinimumSpace := SetupHeader.ExtraDiskSpaceRequired;

    for I := 0 to Entries[seFile].Count-1 do begin
      with PSetupFileEntry(Entries[seFile][I])^ do begin
        if LocationEntry <> -1 then begin { not an "external" file }
          if Components = '' then { no types or a file that doesn't belong to any component }
            if (Tasks = '') then { don't count tasks }
              Inc(MinimumSpace, PSetupFileLocationEntry(Entries[seFileLocation][LocationEntry])^.OriginalSize)
        end else begin
          try
            CurFileSize := GetSizeOfFiles(ChangeDirConst(SourceFilename));
          except
            { ignore exceptions }
            CurFileSize := 0;
          end;
          if CurFileSize < 0 then
            CurFileSize := 0;
          ExternalSize := CurFileSize;
          if Components = '' then { no types or a file that doesn't belong to any component }
            if (Tasks = '') then { don't count tasks }
              Inc (MinimumSpace, CurFileSize);
        end;
      end;
    end;

    for I := 0 to Entries[seComponent].Count-1 do
      with PSetupComponentEntry(Entries[seComponent][I])^ do
        Size := GetSizeOfComponent(Name, ExtraDiskSpaceRequired);

    if Entries[seType].Count > 0 then begin
      MinimumTypeSpace := -1;
      for I := 0 to Entries[seType].Count-1 do begin
        with PSetupTypeEntry(Entries[seType][I])^ do begin
          Size := GetSizeOfType(Name, toIsCustom in Options);
          if (MinimumTypeSpace = -1) or (Size < MinimumTypeSpace) then
            MinimumTypeSpace := Size;
        end;
      end;
      Inc(MinimumSpace, MinimumTypeSpace);
    end;

  finally
    SetErrorMode (OldErrorMode);
  end;

end;

procedure DeinitSetup;
var
  I: Integer;
begin
  for I := 0 to DeleteFilesAfterInstallList.Count-1 do
    DeleteFile (DeleteFilesAfterInstallList[I]);
  DeleteFilesAfterInstallList.Clear;
  for I := DeleteDirsAfterInstallList.Count-1 downto 0 do
    RemoveDirectory (PChar(DeleteDirsAfterInstallList[I]));
  DeleteDirsAfterInstallList.Clear;

  { Free the _isbunzp.dll handle }
  if BunzipDLLHandle <> 0 then
    FreeLibrary (BunzipDLLHandle);

  { Free the shfolder.dll handles }
  if SHFolderDLLHandle <> 0 then begin
    @SHGetFolderPathFunc := nil;
    FreeLibrary (SHFolderDLLHandle);
  end;
  if SHFolderShell32Handle <> 0 then
    FreeLibrary (SHFolderShell32Handle);

  if TempInstallDir <> '' then
    DelTree (TempInstallDir, True, True, True, nil, nil);

  EndDebug;

  if RestartSystem then begin
    if SetupLdrMode then
      { Send a special message back to SetupLdr telling it to restart the system
        after Setup returns }
      SendMessage (SetupLdrWnd, WM_USER + 150, 10000, 0)
    else
      { Restart now if we're not being run from SetupLdr.
        Note: RestartComputer doesn't return if successful. }
      RestartComputer;
  end;
end;

function BlendRGB (const Color1, Color2: TColor; const Blend: Integer): TColor;
{ Blends Color1 and Color2. Blend must be between 0 and 63; 0 = all Color1,
  63 = all Color2. }
type
  TColorBytes = array[0..3] of Byte;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to 2 do
    TColorBytes(Result)[I] := Integer(TColorBytes(Color1)[I] +
      ((TColorBytes(Color2)[I] - TColorBytes(Color1)[I]) * Blend) div 63);
end;

function ExitSetupMsgBox: Boolean;
begin
  Result := MsgBox(SetupMessages[msgExitSetupMessage], SetupMessages[msgExitSetupTitle],
    mbConfirmation, MB_YESNO or MB_DEFBUTTON2) = ID_YES;
end;

function SetFontNameSize (const AFont: TFont; const AName: String;
  const ASize: Integer; const ADefaultName: String;
  const ADefaultSize: Integer): Boolean;
{ Returns True if it used the AName as the font name, or False if it used
  ADefaultName as the font name. }
var
  N: String;
  S: Integer;
begin
  N := AName;
  S := ASize;
  Result := FontExists(N);
  if not Result then begin
    N := ADefaultName;
    S := ADefaultSize;
  end;
  if S >= 0 then  { points (positive) }
    S := MulDiv(-S, Screen.PixelsPerInch, 72)
  else  { pixels (negative) }
    S := MulDiv(S, Screen.PixelsPerInch, 96);

  AFont.Name := N;
  AFont.Height := S;
end;

var
  SavedDefaultDialogFontName: String;

function GetDefaultDialogFontName: String;
begin
  if SavedDefaultDialogFontName = '' then begin
    if FontExists('MS Shell Dlg') then
      SavedDefaultDialogFontName := 'MS Shell Dlg'
    else
      { MS Shell Dlg is not available on Japanese & Chinese Windows }
      SavedDefaultDialogFontName := 'MS Sans Serif';
  end;
  Result := SavedDefaultDialogFontName;
end;

type
  TFormAccess = class(TForm);

procedure SetFormFont (Form: TForm; var OldTextHeight, NewTextHeight: Integer);
var
  W, H: Integer;
  R: TRect;
begin
  with Form do begin
    if SetFontNameSize(Font, LangOptions.DialogFontName,
       LangOptions.DialogFontSize, GetDefaultDialogFontName, 8) then
      OldTextHeight := LangOptions.DialogFontStandardHeight
    else
      OldTextHeight := 13;
    NewTextHeight := Canvas.TextHeight('0');
    if NewTextHeight <> OldTextHeight then begin
      { Loosely based on scaling code from TForm.ReadState: }
      TFormAccess(Form).ScaleControls (NewTextHeight, OldTextHeight);
      R := ClientRect;
      W := MulDiv(R.Right, NewTextHeight, OldTextHeight);
      H := MulDiv(R.Bottom, NewTextHeight, OldTextHeight);
      SetBounds (Left, Top, W + (Width - R.Right), H + (Height - R.Bottom));
    end;
  end;
end;


{ TMainForm }

constructor TMainForm.Create (AOwner: TComponent);
var
  SystemMenu: HMenu;
  Palette: packed record
    palVersion: Word;
    palNumEntries: Word;
    palPalEntry: array[0..63] of TPaletteEntry;
  end;
  Entry: Integer;
  C1, C2: TColor;
begin
  inherited;

  if shWindowVisible in SetupHeader.Options then begin
    { Should the main window not be sizable? }
    if not(shWindowShowCaption in SetupHeader.Options) then
      BorderStyle := bsNone
    else
    if not(shWindowResizable in SetupHeader.Options) then
      BorderStyle := bsSingle;

    { Make the main window full-screen }
    SetBounds (0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
    { Before maximizing the window, ensure Handle is created now so the correct
      'restored' position is saved properly }
    HandleNeeded;

    { Maximize the window so that the taskbar is still accessible }
    if shWindowStartMaximized in SetupHeader.Options then
      WindowState := wsMaximized;
  end
  else begin
    Application.ShowMainForm := False;
  end;

  Caption := FmtSetupMessage1(msgSetupWindowTitle, SetupHeader.AppName);

  { Append the 'About Setup' item to the system menu }
  SystemMenu := GetSystemMenu(Handle, False);
  AppendMenu (SystemMenu, MF_SEPARATOR, 0, nil);
  AppendMenu (SystemMenu, MF_STRING, 9999, PChar(SetupMessages[msgAboutSetupMenuItem]));

  if SetupHeader.BackColor <> SetupHeader.BackColor2 then begin
    { Create the palette }
    with Palette do begin
      palVersion := $300;
      palNumEntries := 64;
    end;
    C1 := ColorToRGB(SetupHeader.BackColor);
    C2 := ColorToRGB(SetupHeader.BackColor2);
    for Entry := 0 to 63 do
      { rest of entries }
      TColor(Palette.palPalEntry[Entry]) := BlendRGB(C1, C2, Entry);
    NewPalette := CreatePalette(PLogPalette(@Palette)^);
  end;

  Application.OnActivate := AppOnActivate;
  Application.HookMainWindow (MainWindowHook);

  if Application.ShowMainForm then
    { Show this form now, so that the focus stays on the wizard form that
      ProceedToNextStep (called in the .dpr) shows }
    Visible := True;
end;

destructor TMainForm.Destroy;
begin
  Application.OnActivate := nil;
  Application.UnhookMainWindow (MainWindowHook);
  inherited;
  if NewPalette <> 0 then
    DeleteObject (NewPalette);
end;

procedure TMainForm.CenterForm (AForm: TForm);
var
  R: TRect;
  X, Y: Integer;
begin
  if shWindowVisible in SetupHeader.Options then begin
    R := ClientRect;
    MapWindowPoints (Handle, 0, R, 2);
  end
  else
    R := Rect(0, 0, GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
  X := ((R.Right+R.Left) div 2)-(AForm.Width div 2);
  if X < 0 then X := 0;
  Y := ((R.Bottom+R.Top) div 2)-(AForm.Height div 2);
  if Y < 0 then Y := 0;
  AForm.SetBounds (X, Y, AForm.Width, AForm.Height);
end;

procedure TMainForm.WMSysCommand (var Message: TWMSysCommand);
begin
  if (Message.CmdType = 9999) and not InExecuteLoop then
    ShowAboutBox
  else
    inherited;
end;

function TMainForm.GetPalette: HPALETTE;
begin
  Result := NewPalette;
end;

procedure TMainForm.AppOnActivate(Sender: TObject);
begin
  { This corrects some rare palette problems }
  if NewPalette <> 0 then
    Repaint;
end;

procedure TMainForm.WMEraseBkgnd (var Message: TWMEraseBkgnd);
begin
  { Since the form paints its entire client area in FormPaint, there is
    no need for the VCL to ever erase the client area with the brush color.
    Doing so only slows it down, so this message handler disables that default
    behavior. }
  Message.Result := 0;
end;

procedure TMainForm.FormPaint(Sender: TObject);
var
  C1, C2: TColor;
  CS: TPoint;
  Z: Integer;
  R, R2: TRect;
begin
  with Canvas do begin
    { Draw the blue background }
    if NewPalette <> 0 then
      SelectPalette (Handle, NewPalette, False);
    if SetupHeader.BackColor = SetupHeader.BackColor2 then begin
      Brush.Color := SetupHeader.BackColor;
      FillRect (ClientRect);
    end
    else begin
      C1 := ColorToRGB(SetupHeader.BackColor);
      C2 := ColorToRGB(SetupHeader.BackColor2);
      CS := ClientRect.BottomRight;
      for Z := 0 to 63 do begin
        Brush.Color := BlendRGB(C1, C2, Z) or $02000000;
        if not(shBackColorHorizontal in SetupHeader.Options) then
          FillRect (Rect(0, MulDiv(CS.Y, Z, 63), CS.X, MulDiv(CS.Y, Z+1, 63)))
        else
          FillRect (Rect(MulDiv(CS.X, Z, 63), 0, MulDiv(CS.X, Z+1, 63), CS.Y));
      end;
    end;

    { Draw the application name and copyright }
    SetBkMode (Handle, TRANSPARENT);

    SetFontNameSize (Font, LangOptions.TitleFontName,
      LangOptions.TitleFontSize, 'Arial', 29);
    if IsMultiByteString(SetupHeader.AppName) then
      { Don't use italics on Japanese characters }
      Font.Style := [fsBold]
    else
      Font.Style := [fsBold, fsItalic];
    R := ClientRect;
    InflateRect (R, -8, -8);
    R2 := R;
    OffsetRect (R2, 4, 4);
    Font.Color := clBlack;
    DrawText (Handle, PChar(SetupHeader.AppName), -1, R2, DT_TOP or DT_LEFT or
      DT_WORDBREAK or DT_NOPREFIX);
    Font.Color := clWhite;
    DrawText (Handle, PChar(SetupHeader.AppName), -1, R, DT_TOP or DT_LEFT or
      DT_WORDBREAK or DT_NOPREFIX);

    SetFontNameSize (Font, LangOptions.CopyrightFontName,
      LangOptions.CopyrightFontSize, 'Arial', 8);
    Font.Style := [];
    R := ClientRect;
    InflateRect (R, -6, -6);
    R2 := R;
    DrawText (Handle, PChar(SetupHeader.AppCopyright), -1, R2, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX or DT_CALCRECT);
    R.Top := R.Bottom - (R2.Bottom - R2.Top);
    R2 := R;
    OffsetRect (R2, 1, 1);
    Font.Color := clBlack;
    DrawText (Handle, PChar(SetupHeader.AppCopyright), -1, R2, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX);
    Font.Color := clWhite;
    DrawText (Handle, PChar(SetupHeader.AppCopyright), -1, R, DT_TOP or DT_RIGHT or
      DT_WORDBREAK or DT_NOPREFIX);
  end;
end;

procedure TMainForm.FormResize(Sender: TObject);
begin
  { Needs to redraw the background whenever the form is resized }
  Repaint;
end;

procedure TMainForm.ShowAboutBox;
var
  S: String;
begin
  { Removing the About box or modifying any existing text inside it is a
    violation of the Inno Setup license agreement; see LICENSE.TXT.
    However, adding additional lines to the end of the About box is
    permitted. }
  S := SetupTitle + ' version ' + SetupVersion + SNewLine;
  if SetupTitle <> 'Inno Setup' then
    S := S + (SNewLine + 'Based on Inno Setup' + SNewLine);
  S := S + ('Copyright (C) 1997-2002 Jordan Russell' + SNewLine +
    'Portions Copyright (C) 2000-2002 Martijn Laan' + SNewLine +
    'All rights reserved.' + SNewLine2 +
    'Inno Setup home page:' + SNewLine +
    'http://www.innosetup.com/');
  if SetupMessages[msgAboutSetupNote] <> '' then
    S := S + SNewLine2 + SetupMessages[msgAboutSetupNote];
  MsgBox(S, SetupMessages[msgAboutSetupTitle], mbInformation, MB_OK);
end;

procedure ProcessMessagesProc; far;
begin
  Application.ProcessMessages;
end;

procedure TMainForm.ProceedToNextStep;

  procedure ProcessRunEntry(const RunEntry: PSetupRunEntry; var ErrorCode: Integer);
  var
    S: String;
  begin
    with RunEntry^ do begin
      S := ChangeDirConst(Name);
      if not(roShellExec in Options) then begin
        if (not(roSkipIfDoesntExist in Options) or NewFileExists(S)) and
           not InstExec(S, ChangeDirConst(Parameters),
             ChangeDirConst(WorkingDir), Wait = rwWaitUntilTerminated,
             Wait = rwWaitUntilIdle, ShowCmd, ProcessMessagesProc,
             ErrorCode) then
          MsgBox (FmtSetupMessage1(msgErrorExecutingProgram, S) + SNewLine2 +
            FmtSetupMessage(msgErrorFunctionFailedWithMessage, ['CreateProcess',
              IntToStr(ErrorCode), SysErrorMessage(ErrorCode)]),
            '', mbCriticalError, MB_OK);
      end
      else begin
        if (not(roSkipIfDoesntExist in Options) or FileOrDirExists(S)) and
           not InstShellExec(S, ChangeDirConst(Parameters),
             ChangeDirConst(WorkingDir), ShowCmd, ErrorCode) then
          MsgBox (FmtSetupMessage1(msgErrorExecutingProgram, S) + SNewLine2 +
            FmtSetupMessage(msgErrorFunctionFailed, ['ShellExecute',
              IntToStr(ErrorCode)]),
            '', mbCriticalError, MB_OK);
      end;
    end;
  end;

var
  I, ErrorCode: Integer;
  RunEntry: PSetupRunEntry;
  CheckBox: TCheckBox;
  CheckIfRestartNeeded: Boolean;
  ChecksumBefore, ChecksumAfter: TFROChecksum;
label NextStep;
begin
  NextStep:
  { Finish up the previous step }
  case CurStep of
    csStart: ;
    csWizard: begin
        { See CVS log for Main.pas 1.23 for explanation of
          RemoveBackslashUnlessRoot calls }
        WizardDirValue := RemoveBackslashUnlessRoot(WizardForm.DirEdit.Text);
        WizardGroupValue := RemoveBackslashUnlessRoot(WizardForm.GroupEdit.Text);
        WizardNoIcons := WizardForm.NoIconsCheck.Checked;
        WizardSetupType := WizardForm.GetSetupType();
        WizardForm.GetSelectedComponents(WizardComponents, False);
        WizardForm.GetSelectedTasks(WizardTasks, False);
        if InitSaveInf <> '' then
          SaveInf(InitSaveInf);
        { Make sure that riched*.dll is unloaded before installation is
          started, so that installations can replace it. }
        WizardForm.LicenseMemo.Text := '';
        WizardForm.LicenseMemo.UseRichEdit := False;
        WizardForm.InfoBeforeMemo.Text := '';
        WizardForm.InfoBeforeMemo.UseRichEdit := False;
      end;
    csCopy: begin
        if Entries[seRun].Count <> 0 then begin
          CheckIfRestartNeeded := (shRestartIfNeededByRun in SetupHeader.Options) and
            not NeedsRestart;
          if CheckIfRestartNeeded then
            MakeFileRenameOperationsChecksum (ChecksumBefore);
          { Disable WizardForm during execution of [Run] entries so that it
            produces a nice "beep" if the user tries clicking on it.
            Note: Can't use DisableTaskWindows or else the focus won't
            return to Setup after the programs terminate. }
          WizardForm.Enabled := False;
          try
            InExecuteLoop := True;
            WizardForm.CancelButton.Enabled := False;
            for I := 0 to Entries[seRun].Count-1 do begin
              RunEntry := PSetupRunEntry(Entries[seRun][I]);
              if ShouldProcessRunEntry(WizardComponents, WizardTasks, RunEntry) and
                 not(roPostInstall in RunEntry.Options) then begin
                if RunEntry.StatusMsg <> '' then
                  WizardForm.StatusLabel.Caption := RunEntry.StatusMsg
                else
                  WizardForm.StatusLabel.Caption := SetupMessages[msgStatusRunProgram];
                WizardForm.StatusLabel.Update;
                if roHideWizard in RunEntry.Options then begin
                  if WizardForm.Visible and not HideWizard then begin
                    HideWizard := True;
                    UpdateWizardFormVisibility;
                  end;
                end
                else begin
                  if HideWizard then begin
                    HideWizard := False;
                    UpdateWizardFormVisibility;
                  end;
                end;
                DebugNotifyEntry(seRun, I);
                ProcessRunEntry(RunEntry, ErrorCode);
              end;
            end;
          finally
            InExecuteLoop := False;
            WizardForm.CancelButton.Enabled := True;
            WizardForm.Enabled := True;
            if HideWizard then begin
              HideWizard := False;
              UpdateWizardFormVisibility;
            end;
            if CheckIfRestartNeeded then begin
              MakeFileRenameOperationsChecksum (ChecksumAfter);
              if not CompareFileRenameOperationsChecksums(ChecksumBefore,
                 ChecksumAfter) then
                NeedsRestart := True;
            end;
          end;
          Application.BringToFront;
        end;
      end;
    csFinished: begin
        if not NeedsRestart then begin
          InExecuteLoop := True;
          try
            for I := 0 to RunCheckBoxes.Count-1 do begin
              CheckBox := TCheckBox(RunCheckboxes[I]);
              if CheckBox.Checked then begin
                RunEntry := PSetupRunEntry(CheckBox.Tag);
                ProcessRunEntry(RunEntry, ErrorCode);
              end;
            end;
          finally
            InExecuteLoop := False;
          end;
          RunCheckboxes.Free();
        end else begin
          case SetupHeader.InstallMode of
            imNormal:
              RestartSystem := WizardForm.YesRadio.Checked;
            imSilent:
              RestartSystem := not InitNoRestart and
                (MsgBox(ExpandSetupMessage(msgFinishedRestartMessage), '',
                 mbConfirmation, MB_YESNO) = IDYES);
            imVerySilent:
              RestartSystem := not InitNoRestart;
          end;
        end;
      end;
  end;

  { Proceed to next step }
  Inc (CurStep);

  case CurStep of
    csWizard: begin
          Application.Restore;
          WizardForm := TWizardForm.Create(Application);
          if SetupHeader.InstallMode <> imNormal then begin
            if SetupHeader.InstallMode = imSilent then begin
              WizardForm.CurPage := wpInstalling;
              WizardForm.CurPageChanged;
              {$IFDEF Delphi3orHigher}
              { Hack needed or else the taskbar button isn't activated when
                WindowVisible=no }
              SetActiveWindow(Application.Handle);
              {$ENDIF}
              WizardForm.Show;
            end;
            goto NextStep;
          end else begin
            {$IFDEF Delphi3orHigher}
            { Hack needed or else the taskbar button isn't activated when
              WindowVisible=no }
            SetActiveWindow(Application.Handle);
            {$ENDIF}
            WizardForm.Show;
          end;
        end;
    csCopy: begin
          Application.Restore;
          Update;
          StartInstalling;
        end;
    csFinished: begin
          if NeedsRestart then begin
            with WizardForm do begin
              ChangeFinishedLabel (ExpandSetupMessage(msgFinishedRestartLabel));
              YesRadio.Visible := True;
              NoRadio.Visible := True;
            end;
          end else begin
            RunCheckBoxes := TList.Create();
            WizardForm.CreateRunButtons(WizardComponents, WizardTasks, RunCheckBoxes);
          end;

          if SetupHeader.InstallMode <> imNormal then
            goto NextStep;

          if SetupHeader.InfoAfterText = '' then begin
            if not NeedsRestart and (shDisableFinishedPage in SetupHeader.Options) then
              goto NextStep;  { proceed to 'csTerminate' step }
            WizardForm.CurPage := wpFinished;
          end
          else begin
            { Don't set up the rich-edit InfoAfterMemo until now, so that
              installations can replace riched*.dll }
            WizardForm.InfoAfterMemo.UseRichEdit := True;
            WizardForm.InfoAfterMemo.RTFText := SetupHeader.InfoAfterText;
            WizardForm.CurPage := wpInfoAfter;
          end;
          Application.Restore;
          Update;
          WizardForm.CurPageChanged;
          WizardForm.Show;
        end;
    csTerminate: Close;
  end;
end;

procedure TMainForm.WMSetupNextStep (var Msg: TMessage);
begin
  try
    ProceedToNextStep;
  except
    { If an exception was raised, display the message, then terminate }
    Application.HandleException (Self);
    SetupExitCode := ecNextStepError;
    Application.Terminate;
  end;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := False;
  if not InExecuteLoop then begin
    if CurStep < csFinished then begin
      if (CurStep = csWizard) and (WizardForm.CurPage = wpPreparing) then
        CanClose := True
      else if (CurStep <> csCopy) or not InitNoCancel then begin
        CanClose := ExitSetupMsgBox;
        if CanClose and (CurStep = csCopy) then
          NeedToAbortInstall := True;
      end;
    end
    else
      CanClose := True;
  end;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if CurStep < csFinished then
    SetupExitCode := ecCancelledBeforeInstall;
end;

procedure TMainForm.WMGetDlgCode (var Message: TWMGetDlgCode);
begin
  Message.Result := Message.Result or DLGC_WANTTAB;
end;

function EWP (Wnd: HWND; Param: Longint): BOOL; stdcall;
begin
  { Note: GetParent is not used here because the other windows are not
    actually child windows since they don't have WS_CHILD set. }
  if GetWindowLong(Wnd, GWL_HWNDPARENT) <> Param then
    Result := True
  else begin
    Result := False;
    BringWindowToTop (Wnd);
  end;
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  { If, for some reason, the user doesn't have a mouse and the main form was
    activated, there would normally be no way to reactivate the child form.
    But this reactivates the form if the user hits a key on the keyboard }
  Key := 0;
  EnumWindows (@EWP, Handle);
end;

procedure TMainForm.UpdateWizardFormVisibility;
var
  ShouldShow: Boolean;
begin
  { Note: We don't adjust WizardForm.Visible because on Delphi 3+, if all forms
    have Visible set to False, the application taskbar button disappears. }
  if Assigned(WizardForm) and WizardForm.HandleAllocated then begin
    ShouldShow := WizardForm.Showing and not HideWizard and
      not IsIconic(Application.Handle);
    if (GetWindowLong(WizardForm.Handle, GWL_STYLE) and WS_VISIBLE <> 0) <> ShouldShow then begin
      if ShouldShow then
        ShowWindow(WizardForm.Handle, SW_SHOW)
      else
        ShowWindow(WizardForm.Handle, SW_HIDE);
    end;
  end;
end;

function TMainForm.MainWindowHook (var Message: TMessage): Boolean;
var
  IsIcon: Boolean;
begin
  Result := False;
  case Message.Msg of
    WM_WINDOWPOSCHANGED: begin
        { When the application window is minimized or restored, also hide or
          show WizardForm.
          Note: MainForm is hidden/shown automatically because its owner
          window is Application.Handle. }
        IsIcon := IsIconic(Application.Handle);
        if IsMinimized <> IsIcon then begin
          IsMinimized := IsIcon;
          UpdateWizardFormVisibility;
        end;
      end;
  end;
end;

procedure TMainForm.WMShowWindow (var Message: TWMShowWindow);
begin
  inherited;
  { When showing, ensure WizardForm is the active window, not MainForm }
  if Message.Show and (GetActiveWindow = Handle) and
     Assigned(WizardForm) and WizardForm.HandleAllocated and
     IsWindowVisible(WizardForm.Handle) then
    SetActiveWindow(WizardForm.Handle);
end;


procedure InitWindowsVersion;

  procedure ReadServicePackFromRegistry;
  var
    K: HKEY;
    Size, Typ, SP: DWORD;
  begin
    if RegOpenKeyEx(HKEY_LOCAL_MACHINE, 'System\CurrentControlSet\Control\Windows',
       0, KEY_QUERY_VALUE, K) = ERROR_SUCCESS then begin
      Size := SizeOf(SP);
      if (RegQueryValueEx(K, 'CSDVersion', nil, @Typ, @SP, @Size) = ERROR_SUCCESS) and
         (Typ = REG_DWORD) and (Size = SizeOf(SP)) then
        NTServicePackLevel := Word(SP);
      RegCloseKey (K);
    end;
  end;

type
  TOSVersionInfoEx = packed record
    dwOSVersionInfoSize: DWORD;
    dwMajorVersion: DWORD;
    dwMinorVersion: DWORD;
    dwBuildNumber: DWORD;
    dwPlatformId: DWORD;
    szCSDVersion: array[0..127] of Char;
    wServicePackMajor: Word;
    wServicePackMinor: Word;
    wSuiteMask: Word;
    wProductType: Byte;
    wReserved: Byte;
  end;
var
  OSVersionInfo: TOSVersionInfo;
  OSVersionInfoEx: TOSVersionInfoEx;
begin
  OSVersionInfo.dwOSVersionInfoSize := SizeOf(OSVersionInfo);
  if GetVersionEx(OSVersionInfo) then begin
    WindowsVersion := (Byte(OSVersionInfo.dwMajorVersion) shl 24) or
      (Byte(OSVersionInfo.dwMinorVersion) shl 16) or
      Word(OSVersionInfo.dwBuildNumber);
    { ^ Note: We MUST clip dwBuildNumber to 16 bits for Win9x compatibility }
    if IsNT then begin
      if OSVersionInfo.dwMajorVersion >= 5 then begin
        { OSVERSIONINFOEX is only available starting in Windows 2000 }
        OSVersionInfoEx.dwOSVersionInfoSize := SizeOf(OSVersionInfoEx);
        if GetVersionEx(POSVersionInfo(@OSVersionInfoEx)^) then
          NTServicePackLevel := (Byte(OSVersionInfoEx.wServicePackMajor) shl 8) or
            Byte(OSVersionInfoEx.wServicePackMinor);
      end
      else if OSVersionInfo.dwMajorVersion = 4 then begin
        { Read from the registry on NT 4 }
        ReadServicePackFromRegistry;
      end;
    end;
  end;
end;

procedure CreateEntryLists;
var
  I: TEntryType;
begin
  for I := Low(I) to High(I) do
    Entries[I] := TList.Create;
end;

procedure FreeEntryLists;
var
  I: TEntryType;
  J: Integer;
  L: TList;
begin
  for I := High(I) downto Low(I) do begin
    L := Entries[I];
    if Assigned(L) then begin
      for J := L.Count-1 downto 0 do begin
        if EntryStrings[I] <> 0 then
          SEFreeRec (L[J], EntryStrings[I])
        else
          FreeMem (L[J]);
        L[J] := nil;
      end;
      L.Free;
      Entries[I] := nil;
    end;
    if Assigned(OriginalEntryIndexes[I]) then begin
      OriginalEntryIndexes[I].Free;
      OriginalEntryIndexes[I] := nil;
    end;
  end;
end;

initialization
  IsNT := UsingWinNT;
  InitWindowsVersion;
  NewGUI := WindowsVersion >= $04000000;
  ConstLeadBytes := @SetupHeader.LeadBytes;
  InitComponents := TStringList.Create();
  WizardComponents := TStringList.Create();
  WizardTasks := TStringList.Create();
  CreateEntryLists;
  DeleteFilesAfterInstallList := TStringList.Create;
  DeleteDirsAfterInstallList := TStringList.Create;

finalization
  WizardImage.Free;
  WizardImage := nil;
  WizardSmallImage.Free;
  WizardSmallImage := nil;

  DeleteDirsAfterInstallList.Free;
  DeleteFilesAfterInstallList.Free;
  FreeEntryLists;
  WizardTasks.Free();
  WizardComponents.Free();
  InitComponents.Free();
end.
