unit NewDisk;

{
  Inno Setup
  Copyright (C) 1998-99 Jordan Russell
  For conditions of distribution and use, see LICENSE.TXT.

  New Disk form
}

interface

{$I VERSION.INC}

uses
  Windows, SysUtils, Messages, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls;

type
  TNewDiskForm = class(TForm)
    DiskPaintBox: TPaintBox;
    SelectDiskLabel: TLabel;
    PathLabel: TLabel;
    PathEdit: TEdit;
    BrowseButton: TButton;
    CancelButton: TButton;
    OKButton: TButton;
    procedure DiskPaintBoxPaint(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure BrowseButtonClick(Sender: TObject);
  private
    { Private declarations }
    Filename: string;
    DiskImage: TBitmap;
  public
    { Public declarations }
    constructor Create (AOwner: TComponent); override;
    destructor Destroy; override;
  end;

function SelectDisk (const DiskNumber: Integer; const AFilename: String; var Path: String): Boolean;

implementation

uses
  ShlObj, {$IFNDEF Delphi3orHigher} Ole2, {$ELSE} ActiveX, {$ENDIF}
  Msgs, MsgIDs, MBCSUtil, CmnFunc, CmnFunc2,
  Main;

{$R *.DFM}

function SelectDisk (const DiskNumber: Integer; const AFilename: String;
  var Path: String): Boolean;
begin
  with TNewDiskForm.Create(Application) do
    try
      Filename := AFilename;
      SelectDiskLabel.Caption := FmtSetupMessage(msgSelectDiskLabel2, [IntToStr(DiskNumber)]);
      PathEdit.Text := Path;
      MessageBeep (0);
      Result := ShowModal = mrOK;
      if Result then
        Path := ExtractFilePath(ExpandFilename(AddBackslash(PathEdit.Text) + Filename));
    finally
      Free;
    end;
end;

function BrowseForFolder (var Directory: String; const ParentWnd: HWND): Boolean;
var
  Malloc: IMalloc;
  BrowseInfo: TBrowseInfo;
  DisplayName, Path: array[0..MAX_PATH-1] of Char;
  IDList: PItemIDList;
begin
  Result := False;
  if FAILED(SHGetMalloc(Malloc)) then
    Malloc := nil;
  FillChar (BrowseInfo, SizeOf(BrowseInfo), 0);
  with BrowseInfo do begin
    hwndOwner := ParentWnd;
    pszDisplayName := @DisplayName;
    lpszTitle := PChar(SetupMessages[msgSelectDirectoryLabel]);
    ulFlags := BIF_RETURNONLYFSDIRS;
  end;
  IDList := SHBrowseForFolder(BrowseInfo);
  try
    if (IDList = nil) or not SHGetPathFromIDList(IDList, Path) then
      Exit;
    Directory := Path;
  finally
    if Assigned(Malloc) then
      Malloc.Free (IDList);
  end;
  Result := True;
end;


{ TNewDiskForm }

constructor TNewDiskForm.Create (AOwner: TComponent);
var
  OldTextHeight, NewTextHeight: Integer;
begin
  inherited;

  SetFormFont (Self, OldTextHeight, NewTextHeight);

  Caption := SetupMessages[msgChangeDiskTitle];
  PathLabel.Caption := SetupMessages[msgPathLabel];
  BrowseButton.Caption := SetupMessages[msgButtonBrowse];
  OKButton.Caption := SetupMessages[msgButtonOK];
  CancelButton.Caption := SetupMessages[msgButtonCancel];

  DiskImage := TBitmap.Create;
  DiskImage.Handle := LoadBitmap(HInstance, 'DISKIMAGE');  {don't localize}
end;

destructor TNewDiskForm.Destroy;
begin
  inherited;
  DiskImage.Free;
end;

procedure TNewDiskForm.DiskPaintBoxPaint(Sender: TObject);
begin
  with DiskPaintBox, Canvas do
    BrushCopy (ClientRect, DiskImage, ClientRect, clBlue);
end;

procedure TNewDiskForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  FN: String;
begin
  case ModalResult of
    mrOK: begin
        FN := ExpandFilename(AddBackslash(Trim(PathEdit.Text)) + Filename);
        if not NewFileExists(FN) then begin
          CanClose := False;
          MsgBox (FmtSetupMessage(msgFileNotInDir2, [Filename,
              RemoveBackslashUnlessRoot(ExtractFilePath(FN))]),
            '', mbError, MB_OK);
        end;
      end;
    mrCancel: CanClose := ExitSetupMsgBox;
  end;
end;

procedure TNewDiskForm.BrowseButtonClick(Sender: TObject);
var
  Dir: String;
begin
  Dir := Trim(PathEdit.Text);
  if BrowseForFolder(Dir, Handle) then
    PathEdit.Text := Dir;
end;

end.
