unit Wizard;

{
  Inno Setup
  Copyright (C) 1998-2002 Jordan Russell
  Portions by Martijn Laan
  For conditions of distribution and use, see LICENSE.TXT.

  Wizard form

  $jrsoftware: issrc/Projects/Wizard.pas,v 1.19 2002/12/16 03:18:04 jr Exp $
}

interface

uses
  Windows, SysUtils, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, NewBevel, NewFCtrl, ExtCtrls,
  ChildFrm, Struct, NewCheckListBox, RichEditViewer, BeveledLabel,
  NewProgressBar, MsgIDs;

type
  TWizardPage = (wpWelcome, wpLicense, wpPassword, wpInfoBefore, wpUserInfo,
    wpSelectDir, wpSelectComponents, wpSelectProgramGroup, wpSelectTasks,
    wpReady, wpPreparing, wpInstalling, wpInfoAfter, wpFinished);

  TWizardForm = class(TSetupChildForm)
    CancelButton: TButton;
    NextButton: TButton;
    BackButton: TButton;
    Notebook1: TNotebook;
    Notebook2: TNotebook;
    DiskSpaceLabel: TLabel;
    DirEdit: TEdit;
    DirList: TNewDirectoryListBox;
    DriveList: TNewDriveComboBox;
    GroupEdit: TEdit;
    NoIconsCheck: TCheckBox;
    GroupList: TListBox;
    PasswordLabel: TLabel;
    PasswordEdit: TEdit;
    PasswordEditLabel: TLabel;
    ReadyMemo: TMemo;
    TypesCombo: TComboBox;
    Bevel: TBevel;
    PicturePaintBox: TPaintBox;
    WelcomeLabel1: TLabel;
    InfoBeforeMemo: TRichEditViewer;
    InfoBeforeClickLabel: TLabel;
    PnlMain: TPanel;
    Bevel1: TBevel;
    PageNameLabel: TLabel;
    PageDescriptionLabel: TLabel;
    SmallPicturePaintBox: TPaintBox;
    ReadyLabel: TLabel;
    FinishedLabel: TLabel;
    YesRadio: TRadioButton;
    NoRadio: TRadioButton;
    PicturePaintBox2: TPaintBox;
    WelcomeLabel2: TLabel;
    LicenseLabel1: TLabel;
    LicenseMemo: TRichEditViewer;
    InfoAfterMemo: TRichEditViewer;
    InfoAfterClickLabel: TLabel;
    ComponentsList: TNewCheckListBox;
    BaseTaskCheckBox: TCheckBox;
    BaseTaskLabel: TLabel;
    ComponentsDiskSpaceLabel: TLabel;
    BeveledLabel: TBeveledLabel;
    StatusLabel: TLabel;
    FilenameLabel: TNewPathLabel;
    ProgressGauge: TNewProgressBar;
    SelectDirLabel: TLabel;
    SelectStartMenuFolderLabel: TLabel;
    SelectComponentsLabel: TLabel;
    SelectTasksLabel: TLabel;
    LicenseAcceptedRadio: TRadioButton;
    LicenseNotAcceptedRadio: TRadioButton;
    UserInfoNameLabel: TLabel;
    UserInfoNameEdit: TEdit;
    UserInfoOrgLabel: TLabel;
    UserInfoOrgEdit: TEdit;
    PreparingErrorPaintBox: TPaintBox;
    PreparingLabel: TLabel;
    FinishedHeadingLabel: TLabel;
    procedure NextButtonClick(Sender: TObject);
    procedure BackButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure DirListChange(Sender: TObject);
    procedure GroupListClick(Sender: TObject);
    procedure NoIconsCheckClick(Sender: TObject);
    procedure PicturePaintBoxPaint(Sender: TObject);
    procedure GroupEditChange(Sender: TObject);
    procedure TypesComboChange(Sender: TObject);
    procedure ComponentsListClickCheck(Sender: TObject);
    procedure ReadyMemoKeyPress(Sender: TObject; var Key: Char);
    procedure FormDestroy(Sender: TObject);
    procedure LicenseAcceptedRadioClick(Sender: TObject);
    procedure LicenseNotAcceptedRadioClick(Sender: TObject);
    procedure PreparingErrorPaintBoxPaint(Sender: TObject);
  private
    { Private declarations }
    OldTextHeight, NewTextHeight: Integer;
    NeedSelectTasksPageUpdate: Boolean;
    FirstRunButton: TButtonControl;
    TaskLabels, TaskButtons: TList;
    PrevAppDir, PrevGroup, PrevSetupType, PrevUserInfoName, PrevUserInfoOrg: String;
    PrevSelectedComponents, PrevDeselectedComponents: TStringList;
    PrevSelectedTasks, PrevDeselectedTasks: TStringList;
    InitialSelectedComponents: TStringList;
    InitialSetupTypeIndex: Integer;
    procedure CalcComponentsSpace;
    function CreateCheckBox(const BaseControl: TControl; const ACaption: String;
      const AChecked: Boolean; const ATag: Integer;
      var Y: Integer): TButtonControl;
    function CreateLabel(const BaseControl: TControl; const ACaption: String;
      var Y: Integer): TLabel;
    function CreateRadioButton(const BaseControl: TControl;
      const ACaption, AGroup: String; const AChecked: Boolean;
      const ATag: Integer; var Y: Integer): TButtonControl;
    procedure CreateTaskButtons(const SelectedComponents: TStringList);
    procedure FindPreviousData;
    function GetFirstTaskButton: TButtonControl;
    function PrepareToInstall: Boolean;
    procedure ReadGroupList;
    procedure WMSysCommand (var Message: TWMSysCommand); message WM_SYSCOMMAND;
  public
    { Public declarations }
    CurPage: TWizardPage;
    constructor Create (AOwner: TComponent); override;
    procedure ChangeReadyLabel (const S: String);
    procedure ChangeFinishedLabel (const S: String);
    procedure CreateRunButtons(const SelectedComponents, SelectedTasks: TStringList; RunButtons: TList);
    function GetSetupType: PSetupTypeEntry;
    procedure SetSelectedComponents(const SelectedComponents, DeselectedComponents: TStringList);
    procedure SetSelectedComponentsFromType(const TypeName: String; OnlySelectFixedComponents: Boolean);
    procedure GetSelectedComponents(Components: TStringList; const Descriptions: Boolean);
    procedure UpdateSelectTasksPage;
    procedure GetSelectedTasks(Tasks: TStringList; const Descriptions: Boolean);
    procedure CurPageChanged;
    function SkipCurPage: Boolean;
  end;

var
  WizardForm: TWizardForm;

function ExpandSetupMessage(const ID: TSetupMessageID): String;

implementation

uses
  ShlObj, GroupRadioButton, Msgs, Main, MBCSUtil, CmnFunc, CmnFunc2,
  zlib, InstFunc;

{$R *.DFM}

const
  NotebookPages: array[TWizardPage, 0..1] of Integer =
    ((0, -1), (1, 0), (1, 1), (1, 2), (1, 3), (1, 4), (1, 5), (1, 6), (1, 7),
     (1, 8), (1, 9), (1, 10), (1, 11), (2, -1));

  PageCaptions: array[wpLicense..wpInfoAfter] of TSetupMessageID =
    (msgWizardLicense, MsgWizardPassword, msgWizardInfoBefore,
     msgWizardUserInfo, msgWizardSelectDir, msgWizardSelectComponents,
     msgWizardSelectProgramGroup, msgWizardSelectTasks, msgWizardReady,
     msgWizardPreparing, msgWizardInstalling, msgWizardInfoAfter);

var
  PageDescriptions: array[wpLicense..wpInfoAfter] of String;
  MinimumComponentsSpace: LongInt;

function IntToKBStr(const I: LongInt): String;
begin
  Result := IntToStr((I + 1023) div 1024);
end;

function IntToMBStr (const I: Longint): String;
var
  MB, MBFrac, MBDec: Longint;
begin
  MB := I div 1048576;
  MBFrac := (I mod 1048576) * 10;
  MBDec := MBFrac div 1048576;
  if MBFrac mod 1048576 <> 0 then  { round up }
    Inc (MBDec);
  if MBDec = 10 then begin
    Inc (MB);
    MBDec := 0;
  end;
  Result := Format('%d%s%d', [MB, DecimalSeparator, MBDec]);
end;

function ExpandSetupMessageEx(const ID: TSetupMessageID;
  const Space: Longint): String;
begin
  Result := SetupMessages[ID];
  {don't localize these}
  StringChange (Result, '[name]', SetupHeader.AppName);
  StringChange (Result, '[name/ver]', SetupHeader.AppVerName);
  StringChange (Result, '[kb]', IntToKBStr(Space));
  StringChange (Result, '[mb]', IntToMBStr(Space));
{
  if MinimumKBSpaceStr = '' then
    MinimumKBSpaceStr := IntToKBStr(MinimumSpace);
  StringChange (Result, '[kb]', MinimumKBSpaceStr);
  if MinimumMBSpaceStr = '' then
    MinimumMBSpaceStr := IntToMBStr(MinimumSpace);
  StringChange (Result, '[mb]', MinimumMBSpaceStr);
}
end;

function ExpandSetupMessage(const ID: TSetupMessageID): String;
begin
  Result := ExpandSetupMessageEx(ID, MinimumSpace);
end;

function RemoveExtraBackslashes (const S: String): String;
var
  I: Integer;
begin
  Result := S;
  I := 1;
  while I < Length(Result) do begin
    if IsLeadByte(Result[I]) then
      Inc (I, 2)
    else if (Result[I] = '\') and (Result[I+1] = '\') then
      Delete (Result, I+1, 1)
    else
      Inc (I);
  end;
end;

function AdjustLabelHeight (const ALabel: TLabel): Integer;
{ Increases or decreases a label's height so that all text fits.
  Returns the difference in height. }
var
  W, H: Integer;
begin
  W := ALabel.Width;
  H := ALabel.Height;
  ALabel.AutoSize := True;
  ALabel.AutoSize := False;
  ALabel.Width := W;
  Result := ALabel.Height - H;
end;

procedure IncTopDecHeight (const AControl: TControl; const Amount: Integer);
begin
  AControl.SetBounds (AControl.Left, AControl.Top + Amount,
    AControl.Width, AControl.Height - Amount);
end;

function GetTextWidth (const DC: HDC; S: String; const Prefix: Boolean): Integer;
{ Returns the width of the specified string using the font currently selected
  into DC. If Prefix is True, it first removes "&" characters as necessary. } 
var
  I: Integer;
  Size: TSize;
begin
  { This procedure is 10x faster than using DrawText with the DT_CALCRECT flag }
  if Prefix then begin
    I := 1;
    while I <= Length(S) do begin
      if not(S[I] in LeadBytes) then begin
        if S[I] = '&' then
          System.Delete (S, I, 1);
        Inc (I);
      end
      else
        Inc (I, 2);
    end;
  end;
  GetTextExtentPoint32 (DC, PChar(S), Length(S), Size);
  Result := Size.cx;
end;

procedure TWizardForm.CalcComponentsSpace();
var
  SelectedComponents: TStringList;
  I: Integer;
begin
  MinimumComponentsSpace := SetupHeader.ExtraDiskSpaceRequired;

  SelectedComponents := TStringList.Create();
  GetSelectedComponents(SelectedComponents, False);

  //we can't simply sum component sizes because of shared files -> add file sizes
  for I := 0 to Entries[seFile].Count-1 do begin
    with PSetupFileEntry(Entries[seFile][I])^ do begin
      if ShouldProcessEntry(SelectedComponents, nil, Components, '') then begin
        if LocationEntry <> -1 then
          Inc(MinimumComponentsSpace, PSetupFileLocationEntry(Entries[seFileLocation][LocationEntry])^.OriginalSize)
        else
          Inc(MinimumComponentsSpace, ExternalSize)
      end;
    end;
  end;

  //don't forget to add extradiskspacerequired values
  for I := 0 to Entries[seComponent].Count-1 do
    with PSetupComponentEntry(Entries[seComponent][I])^ do
      if SelectedComponents.IndexOf(Name) <> -1 then
        Inc(MinimumComponentsSpace, ExtraDiskSpaceRequired);

  SelectedComponents.Free();

  ComponentsDiskSpaceLabel.Caption := ExpandSetupMessageEx(msgComponentsDiskSpaceMBLabel,
    MinimumComponentsSpace);
end;

{ TWizardForm }

constructor TWizardForm.Create (AOwner: TComponent);
{ Do all initialization of the wizard form. We're overriding Create instead of
  using the FormCreate event, because if an exception is raised in FormCreate
  it's not propogated out. }

  function CalculateWizardButtonWidth: Integer;
  const
    ButtonCaptions: array[0..4] of TSetupMessageID =
      (msgButtonBack, msgButtonCancel, msgButtonFinish, msgButtonInstall,
       msgButtonNext);
  var
    DC: HDC;
    I, W: Integer;
  begin
    Result := MulDiv(75, NewTextHeight, OldTextHeight);
    { Increase the button size if there are unusually long button captions }
    DC := GetDC(0);
    try
      SelectObject (DC, Font.Handle);
      for I := Low(ButtonCaptions) to High(ButtonCaptions) do begin
        W := GetTextWidth(DC, SetupMessages[ButtonCaptions[I]], True) + 20;
        if Result < W then
          Result := W;
      end;
    finally
      ReleaseDC (0, DC);
    end;
  end;

  function TrySetDirListDirectory (const S: String): Boolean;
  begin
    Result := True;
    try
      DirList.Directory := S;
    except
      Result := False;
    end;
  end;

var
  X, W1, W2, W3: Integer;
  S: String;
  SystemMenu: HMENU;
  P, P2, P3: String;
  I, PrevSetupTypeIndex: Integer;
  TypeEntry: PSetupTypeEntry;
  ComponentEntry: PSetupComponentEntry;
  MaxComponentSize: Integer;
  SaveClientWidth, SaveClientHeight: Integer;
begin
  inherited;

  InitialSelectedComponents := TStringList.Create();
  PrevSelectedComponents := TStringList.Create();
  PrevDeselectedComponents := TStringList.Create();
  PrevSelectedTasks := TStringList.Create();
  PrevDeselectedTasks := TStringList.Create();
  TaskLabels := TList.Create();
  TaskButtons := TList.Create();

  SetFormFont (Self, OldTextHeight, NewTextHeight);
  SetFontNameSize (WelcomeLabel1.Font, LangOptions.WelcomeFontName,
    LangOptions.WelcomeFontSize, GetDefaultDialogFontName, 12);
  WelcomeLabel1.Font.Style := [fsBold];
  PageNameLabel.Font.Style := [fsBold];

  if shWindowVisible in SetupHeader.Options then
    Caption := SetupMessages[msgSetupAppTitle]
  else
    Caption := FmtSetupMessage1(msgSetupWindowTitle, SetupHeader.AppName);

  { Give it a minimize button if main window isn't visible }
  if not(shWindowVisible in SetupHeader.Options) then begin
    { Save ClientWidth/ClientHeight and restore them after changing
      BorderStyle. Needed for NT 3.x. }
    SaveClientWidth := ClientWidth;
    SaveClientHeight := ClientHeight;
    BorderIcons := BorderIcons + [biMinimize];
    BorderStyle := bsSingle;
    ClientWidth := SaveClientWidth;
    ClientHeight := SaveClientHeight;
  end;

  { Position the buttons, and scale their size }
  W1 := CalculateWizardButtonWidth;  { width of each button }
  W2 := MulDiv(12, NewTextHeight, OldTextHeight);
  W3 := MulDiv(10, NewTextHeight, OldTextHeight);  { space between Next and Cancel button }

  BackButton.Width := W1;
  NextButton.Width := W1;
  CancelButton.Width := W1;
  X := ClientWidth - W2 - W1;
  CancelButton.Left := X;
  Dec (X, W3);
  Dec (X, W1);
  NextButton.Left := X;
  Dec (X, W1);
  BackButton.Left := X;

  { Initialize wpWelcome page }
  WelcomeLabel1.Caption := ExpandSetupMessage(msgWelcomeLabel1) + SNewLine;
  AdjustLabelHeight(WelcomeLabel1);
  IncTopDecHeight(WelcomeLabel2, (WelcomeLabel1.Top + WelcomeLabel1.Height) -
    WelcomeLabel2.Top);
  WelcomeLabel2.Caption := ExpandSetupMessage(msgWelcomeLabel2) + SNewLine2 +
    SetupMessages[msgClickNext];

  { Initialize wpLicense page }
  LicenseLabel1.Caption := ExpandSetupMessage(msgLicenseLabel3);
  I := AdjustLabelHeight(LicenseLabel1);
  IncTopDecHeight (LicenseMemo, I);
  LicenseAcceptedRadio.Caption := SetupMessages[msgLicenseAccepted];
  LicenseNotAcceptedRadio.Caption := SetupMessages[msgLicenseNotAccepted];

  { Initialize wpPassword page }
  PasswordLabel.Caption := SetupMessages[msgPasswordLabel3];
  PasswordEditLabel.Caption := SetupMessages[msgPasswordEditLabel];
  I := AdjustLabelHeight(PasswordLabel);
  PasswordEditLabel.Top := PasswordEditLabel.Top + I;
  PasswordEdit.Top := PasswordEdit.Top + I;

  { Initialize wpInfoBefore page }
  InfoBeforeClickLabel.Caption := SetupMessages[msgInfoBeforeClickLabel];
  I := AdjustLabelHeight(InfoBeforeClickLabel);
  IncTopDecHeight (InfoBeforeMemo, I);

  { Initialize wpUserInfo page }
  UserInfoNameLabel.Caption := SetupMessages[msgUserInfoName];
  I := AdjustLabelHeight(UserInfoNameLabel);
  UserInfoNameEdit.Top := UserInfoNameEdit.Top + I;
  UserInfoOrgLabel.Top := UserInfoOrgLabel.Top + I;
  UserInfoOrgLabel.Caption := SetupMessages[msgUserInfoOrg];
  Inc(I, AdjustLabelHeight(UserInfoOrgLabel));
  UserInfoOrgEdit.Top := UserInfoOrgEdit.Top + I;

  { Initialize wpSelectDir page }
  SelectDirLabel.Caption := ExpandSetupMessage(msgSelectDirLabel);
  I := AdjustLabelHeight(SelectDirLabel);
  DirEdit.Top := DirEdit.Top + I;
  IncTopDecHeight (DirList, I);
  DiskSpaceLabel.Caption := ExpandSetupMessage(msgDiskSpaceMBLabel);

  { Initialize wpSelectComponents page }
  SelectComponentsLabel.Caption := ExpandSetupMessage(msgSelectComponentsLabel2);
  I := AdjustLabelHeight(SelectComponentsLabel);
  TypesCombo.Top := TypesCombo.Top + I;
  IncTopDecHeight (ComponentsList, I);
  ComponentsDiskSpaceLabel.Caption := ExpandSetupMessage(msgComponentsDiskSpaceMBLabel);

  { Initialize wpSelectProgramGroup page }
  SelectStartMenuFolderLabel.Caption := ExpandSetupMessage(msgSelectStartMenuFolderLabel);
  I := AdjustLabelHeight(SelectStartMenuFolderLabel);
  GroupEdit.Top := GroupEdit.Top + I;
  IncTopDecHeight (GroupList, I);
  if HasTaskIcons then
    NoIconsCheck.Caption := SetupMessages[msgNoProgramGroupCheck2]
  else
    NoIconsCheck.Caption := SetupMessages[msgNoIconsCheck];

  { Initialize wpSelectTasks page }
  SelectTasksLabel.Caption := ExpandSetupMessage(msgSelectTasksLabel2);
  I := AdjustLabelHeight(SelectTasksLabel);
  BaseTaskLabel.Top := BaseTaskLabel.Top + I;
  BaseTaskCheckBox.Top := BaseTaskCheckBox.Top + I;

  { Initialize wpInfoAfter page }
  InfoAfterClickLabel.Caption := SetupMessages[msgInfoAfterClickLabel];
  I := AdjustLabelHeight(InfoAfterClickLabel);
  IncTopDecHeight (InfoAfterMemo, I);

  { Initialize wpFinished page }
  SetFontNameSize(FinishedHeadingLabel.Font, LangOptions.WelcomeFontName,
    LangOptions.WelcomeFontSize, GetDefaultDialogFontName, 12);
  FinishedHeadingLabel.Font.Style := [fsBold];
  FinishedHeadingLabel.Caption := ExpandSetupMessage(msgFinishedHeadingLabel) +
    SNewLine;
  AdjustLabelHeight(FinishedHeadingLabel);
  FinishedLabel.Top := FinishedHeadingLabel.Top + FinishedHeadingLabel.Height;
  if HasIcons then
    S := ExpandSetupMessage(msgFinishedLabel)
  else
    S := ExpandSetupMessage(msgFinishedLabelNoIcons);
  ChangeFinishedLabel (S + SNewLine2 + SetupMessages[msgClickFinish]);
  FirstRunButton := nil;
  YesRadio.Caption := SetupMessages[msgYesRadio];
  NoRadio.Caption := SetupMessages[msgNoRadio];

  { Set page descriptions }
  PageDescriptions[wpLicense] := SetupMessages[msgLicenseLabel];
  PageDescriptions[wpPassword] := SetupMessages[msgPasswordLabel1];
  PageDescriptions[wpInfoBefore] := SetupMessages[msgInfoBeforeLabel];
  PageDescriptions[wpUserInfo] := SetupMessages[msgUserInfoDesc];
  PageDescriptions[wpSelectDir] := ExpandSetupMessage(msgSelectDirDesc);
  PageDescriptions[wpSelectComponents] := ExpandSetupMessage(msgSelectComponentsDesc);
  PageDescriptions[wpSelectProgramGroup] := ExpandSetupMessage(msgSelectStartMenuFolderDesc);
  PageDescriptions[wpSelectTasks] := ExpandSetupMessage(msgSelectTasksDesc);
  PageDescriptions[wpReady] := ExpandSetupMessage(msgReadyLabel1);
  PageDescriptions[wpPreparing] := ExpandSetupMessage(msgPreparingDesc);
  PageDescriptions[wpInstalling] := ExpandSetupMessage(msgInstallingLabel);
  PageDescriptions[wpInfoAfter] := SetupMessages[msgInfoAfterLabel];

  { Initialize BeveledLabel }
  BeveledLabel.Caption := ' ' + SetupMessages[msgBeveledLabel];

  { Don't set UseRichEdit to True on the TRichEditViewers unless they are going
    to be used. There's no need to load riched*.dll unnecessarily. }
  if SetupHeader.LicenseText <> '' then begin
    LicenseMemo.UseRichEdit := True;
    LicenseMemo.RTFText := SetupHeader.LicenseText;
  end;
  if SetupHeader.InfoBeforeText <> '' then begin
    InfoBeforeMemo.UseRichEdit := True;
    InfoBeforeMemo.RTFText := SetupHeader.InfoBeforeText;
  end;
  { Note: InfoAfterMemo is set up in the Main unit }

  { Center the form inside the MainForm's client area }
  MainForm.CenterForm (Self);

  { Append an 'About Setup' item to the wizard form also }
  SystemMenu := GetSystemMenu(Handle, False);
  AppendMenu (SystemMenu, MF_SEPARATOR, 0, nil);
  AppendMenu (SystemMenu, MF_STRING, 9999, PChar(SetupMessages[msgAboutSetupMenuItem]));

  { Read settings from a previous install if available }
  FindPreviousData;
  PrevSetupTypeIndex := -1; //assigned later

  { Assign default user name & organization on User Info page }
  if shUserInfoPage in SetupHeader.Options then begin
    if PrevUserInfoName = '' then begin
      UserInfoNameEdit.Text := ChangeDirConst(SetupHeader.DefaultUserInfoName);
      UserInfoOrgEdit.Text := ChangeDirConst(SetupHeader.DefaultUserInfoOrg);
    end
    else begin
      UserInfoNameEdit.Text := PrevUserInfoName;
      UserInfoOrgEdit.Text := PrevUserInfoOrg;
    end;
  end;

  { Assign default directory name }
  if shCreateAppDir in SetupHeader.Options then begin
    { Try different values and see what works. The 'except end;' throws away any
      exceptions that might occur. }
    TrySetDirListDirectory ('\');
    TrySetDirListDirectory (SystemDrive + '\');
    P := PrevAppDir;
    if P = '' then
      P := ChangeDirConst(SetupHeader.DefaultDirName);
    if not(shDisableAppendDir in SetupHeader.Options) or
       not TrySetDirListDirectory(P) then begin
      P2 := P;
      while True do begin
        P3 := RemoveBackslashUnlessRoot(ExtractFilePath(P2));
        if Length(P3) < 3 then Break;
        if TrySetDirListDirectory(P3) or (P3 = P2) then
          Break;
        P2 := P3;
      end;
    end;
    DirEdit.Text := P;

    if (InitDir <> '') and not(shDisableDirPage in SetupHeader.Options) then begin
      { ^ InitDir currently isn't supported for installations with
        shDisableDirPage set. If the wizard page isn't displayed, it doesn't
        get a chance to validate the path specified. }
      try
        DirList.Directory := ExtractFilePath(InitDir);
        DirList.Directory := InitDir;
      except
      end;
      DirEdit.Text := InitDir;
    end;
  end
  else
    DirEdit.Text := WinDir;

  { Fill types list and assign default type }
  if Entries[seType].Count > 0 then begin
    //first fill list
    TypesCombo.Clear();
    for I := 0 to Entries[seType].Count-1 do begin
      TypeEntry := PSetupTypeEntry(Entries[seType][I]);
      TypesCombo.Items.AddObject(ChangeDirConst(TypeEntry.Description), TObject(TypeEntry));
      if Comparetext(TypeEntry.Name, PrevSetupType) = 0 then
        PrevSetupTypeIndex := I;
    end;
    //now assign default type
    if PrevSetupTypeIndex <> -1 then
      TypesCombo.ItemIndex := PrevSetupTypeIndex
    else
      TypesCombo.ItemIndex := 0;
  end;

  { Fill components list and assign default components}
  //find the biggest component
  MaxComponentSize := 0;
  for I := 0 to Entries[seComponent].Count-1 do begin
    ComponentEntry := PSetupComponentEntry(Entries[seComponent][I]);
    if ComponentEntry.Size > MaxComponentSize then
      MaxComponentSize := ComponentEntry.Size;
  end;

  //first fill list
  ComponentsList.Clear();
  ComponentsList.Flat := shFlatComponentsList in SetupHeader.Options;
  for I := 0 to Entries[seComponent].Count-1 do begin
    ComponentEntry := PSetupComponentEntry(Entries[seComponent][I]);
    ComponentsList.Items.Add(ChangeDirConst(ComponentEntry.Description));
    ComponentsList.ItemObject[I] := TObject(ComponentEntry);
    ComponentsList.ItemEnabled[I] := not (coFixed in ComponentEntry.Options);
    if (shShowComponentSizes in SetupHeader.Options) then begin
      if MaxComponentSize < 1024*1024 then
        ComponentsList.ItemSubItem[I] := FmtSetupMessage1(msgComponentSize1, IntToKBStr(ComponentEntry.Size))
      else
        ComponentsList.ItemSubItem[I] := FmtSetupMessage1(msgComponentSize2, IntToMBStr(ComponentEntry.Size));
    end;
  end;

  //now assign default components
  if InitComponents.Count <> 0 then begin
    SetSelectedComponents(InitComponents, nil);
    //try to find a custom type. if there's no custom type the UsePreviousSetupType won't
    //work and if the install is not silent the type list won't display a custom type.
    if HasCustomType then begin
      for I := 0 to Entries[seType].Count-1 do begin
        TypeEntry := PSetupTypeEntry(Entries[seType][I]);
        if toIsCustom in TypeEntry.Options then begin
          TypesCombo.ItemIndex := I;
          Break;
        end;
      end;
    end;
  end else begin
    if PrevSetupTypeIndex <> -1 then begin
      TypeEntry := PSetupTypeEntry(Entries[seType][PrevSetupTypeIndex]);
      if toIsCustom in TypeEntry.Options then begin
        //the previous setup type is a custom type: first select the default components
        //for the default type (usually the full type). needed for new components.
        TypeEntry := PSetupTypeEntry(Entries[seType][0]);
        SetSelectedComponentsFromType(TypeEntry.Name, False);
        //now restore the customization
        SetSelectedComponents(PrevSelectedComponents, PrevDeselectedComponents);
      end else begin
        //this is not a custom type, so just select components based on the previous type
        TypeEntry := PSetupTypeEntry(Entries[seType][PrevSetupTypeIndex]);
        SetSelectedComponentsFromType(TypeEntry.Name, False);
      end;
    end else if Entries[seType].Count > 0 then begin
      TypeEntry := PSetupTypeEntry(Entries[seType][0]);
      SetSelectedComponentsFromType(TypeEntry.Name, False);
    end;
  end;

  CalcComponentsSpace();

  //Show or hide the components list based on the selected type
  if HasCustomType then begin
    TypeEntry := PSetupTypeEntry(Entries[seType][TypesCombo.ItemIndex]);
    if (toIsCustom in TypeEntry.Options) or (shAlwaysShowComponentsList in SetupHeader.Options) then
      ComponentsList.Visible := True
    else
      ComponentsList.Visible := False;
  end else
    ComponentsList.Visible := False;
  ComponentsDiskSpaceLabel.Visible := ComponentsList.Visible;

  //Store the initial setup type and components (only necessary if customizable)
  if HasCustomType then begin
    InitialSetupTypeIndex := TypesCombo.ItemIndex;
    GetSelectedComponents(InitialSelectedComponents, False);
  end;

  { Assign default group name }
  if (InitProgramGroup <> '') and not(shDisableProgramGroupPage in SetupHeader.Options) then
    { ^ InitProgramGroup currently isn't supported for installations with
      shDisableProgramGroupPage set. If the wizard page isn't displayed, it
      doesn't get a chance to validate the program group name specified. }
    GroupEdit.Text := InitProgramGroup
  else begin
    if (PrevGroup = '') or (PrevGroup = '(Default)') then
      GroupEdit.Text := ChangeDirConst(SetupHeader.DefaultGroupName)
    else
      GroupEdit.Text := PrevGroup;
  end;

  if shAllowNoIcons in SetupHeader.Options then begin
    if InitNoIcons then
      NoIconsCheck.Checked := True;
    NoIconsCheck.Visible := True;
  end else
    NoIconsCheck.Visible := False;

  if SetupHeader.InstallMode = imNormal then begin
    //tasks page will be updated later based on the selected components
    NeedSelectTasksPageUpdate := True;
  end else begin
    //update tasks page now
    UpdateSelectTasksPage();
    NeedSelectTasksPageUpdate := False; //not really necessary
  end;

  CurPage := Low(TWizardPage);
  CurPageChanged;

  ReadGroupList;
end;

procedure TWizardForm.FormDestroy(Sender: TObject);
begin
  TaskButtons.Free();
  TaskLabels.Free();
  PrevDeselectedComponents.Free();
  PrevSelectedTasks.Free();
  PrevDeselectedTasks.Free();
  PrevSelectedComponents.Free();
  InitialSelectedComponents.Free();
end;

procedure TWizardForm.FindPreviousData;
const
  RootKeys: array[0..1] of HKEY = (HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE);
var
  I: Integer;
  H: HKEY;
  S: String;
begin
  { First look in HKEY_CURRENT_USER. That's where Inno Setup creates
    the uninstall key for non-administrators. If the key doesn't exist
    under HKEY_CURRENT_USER, check HKEY_LOCAL_MACHINE. }
  for I := 0 to 1 do
    if RegOpenKeyEx(RootKeys[I],
       PChar(Format('%s\%s_is1', [NEWREGSTR_PATH_UNINSTALL, UninstallRegKeyBaseName])),
       0, KEY_QUERY_VALUE, H) = ERROR_SUCCESS then begin
      try
        { do not localize or change the following strings }
        if shUsePreviousAppDir in SetupHeader.Options then
          RegQueryStringValue (H, 'Inno Setup: App Path', PrevAppDir);
        if shUsePreviousGroup in SetupHeader.Options then
          RegQueryStringValue (H, 'Inno Setup: Icon Group', PrevGroup);
        if shUsePreviousSetupType in SetupHeader.Options then begin
          RegQueryStringValue (H, 'Inno Setup: Setup Type', PrevSetupType);
          if RegQueryStringValue (H, 'Inno Setup: Selected Components', S) then
            PrevSelectedComponents.CommaText := S;
          if RegQueryStringValue (H, 'Inno Setup: Deselected Components', S) then
            PrevDeselectedComponents.CommaText := S;
        end;
        if shUsePreviousTasks in SetupHeader.Options then begin
          if RegQueryStringValue (H, 'Inno Setup: Selected Tasks', S) then
            PrevSelectedTasks.CommaText := S;
          if RegQueryStringValue (H, 'Inno Setup: Deselected Tasks', S) then
            PrevDeselectedTasks.CommaText := S;
        end;
        if shUsePreviousUserInfo in SetupHeader.Options then begin
          RegQueryStringValue (H, 'Inno Setup: User Info: Name', PrevUserInfoName);
          RegQueryStringValue (H, 'Inno Setup: User Info: Organization', PrevUserInfoOrg);
        end;
      finally
        RegCloseKey (H);
      end;
      Break;
    end;
end;

procedure TWizardForm.ChangeReadyLabel (const S: String);
var
  Width, OldHeight, dY: Integer;
begin
  Width := ReadyLabel.Width;
  OldHeight := ReadyLabel.Height;
  ReadyLabel.Caption := S;
  ReadyLabel.AutoSize := True;
  ReadyLabel.Width := Width;
  dY := ReadyLabel.Height-OldHeight;
  ReadyMemo.Top := ReadyMemo.Top+dY;
  ReadyMemo.Height := ReadyMemo.Height-dY;
end;

procedure TWizardForm.ChangeFinishedLabel (const S: String);
var
  Y: Integer;
begin
  FinishedLabel.Caption := S + SNewLine;
  AdjustLabelHeight(FinishedLabel);
  Y := FinishedLabel.Top + FinishedLabel.Height;
  YesRadio.Top := Y;
  NoRadio.Top := Y + MulDiv(24, NewTextHeight, OldTextHeight);
end;

function TWizardForm.CreateLabel(const BaseControl: TControl;
  const ACaption: String; var Y: Integer): TLabel;
var
  ALabel: TLabel;
begin
  ALabel := TLabel.Create(BaseControl.Owner);
  with ALabel do begin
    Parent := BaseControl.Parent;
    AutoSize := False;
    Caption := ACaption;
    Left := BaseControl.Left;
    Top := Y;
    Width := BaseControl.Width;
    Height := BaseControl.Height;
    Y := Y + MulDiv(20, NewTextHeight, OldTextHeight);
  end;

  Result := ALabel;
end;

function TWizardForm.CreateCheckBox(const BaseControl: TControl;
  const ACaption: String; const AChecked: Boolean; const ATag: Integer;
  var Y: Integer): TButtonControl;
var
  ACheckBox: TCheckBox;
begin
  ACheckBox := TCheckBox.Create(BaseControl.Owner);
  with ACheckBox do begin
    Parent := BaseControl.Parent;
    Caption := ACaption;
    Left := BaseControl.Left;
    Top := Y;
    Width := BaseControl.Width;
    Height := BaseControl.Height;
    Checked := AChecked;
    Tag := ATag;
    Y := Y + MulDiv(24, NewTextHeight, OldTextHeight);
  end;

  Result := ACheckBox;
end;

function TWizardForm.CreateRadioButton(const BaseControl: TControl;
  const ACaption, AGroup: String; const AChecked: Boolean; const ATag: Integer;
  var Y: Integer): TButtonControl;
var
  ARadioButton: TGroupRadioButton;
begin
  ARadioButton := TGroupRadioButton.Create(BaseControl.Owner);
  with ARadioButton do begin
    Parent := BaseControl.Parent;
    Caption := ACaption;
    Group := AGroup;
    Left := BaseControl.Left;
    Top := Y;
    Width := BaseControl.Width;
    Height := BaseControl.Height;
    Checked := AChecked;
    Tag := ATag;
    Y := Y + MulDiv(24, NewTextHeight, OldTextHeight);
  end;

  Result := ARadioButton;
end;

procedure TWizardForm.CreateRunButtons(const SelectedComponents, SelectedTasks: TStringList; RunButtons: TList);
var
  RunEntry: PSetupRunEntry;
  RunButton: TButtonControl;
  Caption: String;
  I, Y: Integer;
begin
  Y := YesRadio.Top;

  for I := 0 to Entries[seRun].Count-1 do begin
    RunEntry := PSetupRunEntry(Entries[seRun][I]);
    if ShouldProcessRunEntry(SelectedComponents, SelectedTasks, RunEntry) and (roPostInstall in RunEntry.Options) then begin
      if RunEntry.Description <> '' then
        Caption := RunEntry.Description
      else if not(roShellExec in RunEntry.Options) then
        Caption := FmtSetupMessage1(msgRunEntryExec, ExtractFileName(RunEntry.Name))
      else
        Caption := FmtSetupMessage1(msgRunEntryShellExec, ExtractFileName(RunEntry.Name));
      RunButton := CreateCheckBox(YesRadio, Caption, not(roUnchecked in RunEntry.Options), Integer(RunEntry), Y);
      RunCheckBoxes.Add(RunButton);
      if FirstRunButton = nil then
        FirstRunButton := RunButton;
    end;
  end;
end;

procedure TWizardForm.CreateTaskButtons(const SelectedComponents: TStringList);

  function MultipleRadioButtonsInGroup(const SelectedComponents: TStringList; const GroupDescription: String): Boolean;
  var
    TaskEntry: PSetupTaskEntry;
    N, I: Integer;
  begin
    N := 0;
    for I := 0 to Entries[seTask].Count-1 do begin
      TaskEntry := PSetupTaskEntry(Entries[seTask][I]);
      if ShouldProcessEntry(SelectedComponents, nil, TaskEntry.Components, '') and (ChangeDirConst(TaskEntry.GroupDescription) = GroupDescription) and (toExclusive in TaskEntry.Options) then begin
        Inc(N);
        if N > 1 then begin
          Result := True;
          Exit;
        end;
      end;
    end;
    Result := False;
  end;

var
  TaskEntry: PSetupTaskEntry;
  TaskLabel: TLabel;
  TaskButton: TButtonControl;
  I, P, Y: Integer;
  Description, GroupDescription: String;
  LastGroupDescription: String;
  UseRadioButtons: Boolean;
begin
  for I := 0 to TaskLabels.Count-1 do
    TLabel(TaskLabels[I]).Free();
  TaskLabels.Clear();

  for I := 0 to TaskButtons.Count-1 do
    TButtonControl(TaskButtons[I]).Free();
  TaskButtons.Clear();

  Y := BaseTaskLabel.Top;
  UseRadioButtons := MultipleRadioButtonsInGroup(SelectedComponents, '');
  LastGroupDescription := '';

  //Create the task buttons with their default checked states
  for I := 0 to Entries[seTask].Count-1 do begin
    TaskEntry := PSetupTaskEntry(Entries[seTask][I]);
    if ShouldProcessEntry(SelectedComponents, nil, TaskEntry.Components, '') then begin
      Description := ChangeDirConst(TaskEntry.Description);
      GroupDescription := ChangeDirConst(TaskEntry.GroupDescription);

      //see if we should add a group label
      if GroupDescription <> LastGroupDescription then begin
        TaskLabel := CreateLabel(BaseTaskLabel, GroupDescription, Y);
        TaskLabels.Add(TaskLabel);
        UseRadioButtons := MultipleRadioButtonsInGroup(SelectedComponents, GroupDescription);
        LastGroupDescription := GroupDescription;
      end;

      //create a checkbox or radiobutton. only create a radiobutton if the group
      //has multiple radiobuttons based on the selected components
      if UseRadioButtons and (toExclusive in TaskEntry.Options) then
        TaskButton := CreateRadioButton(BaseTaskCheckBox, Description, GroupDescription, not(toUnchecked in TaskEntry.Options), Integer(TaskEntry), Y)
      else
        TaskButton := CreateCheckBox(BaseTaskCheckBox, Description, not(toUnchecked in TaskEntry.Options), Integer(TaskEntry), Y);

      TaskButtons.Add(TaskButton);
    end;
  end;

  //Now restore the previous checked state of the buttons we just created
  for I := 0 to TaskButtons.Count-1 do begin
    TaskButton := TaskButtons[I];
    TaskEntry := PSetupTaskEntry(TaskButton.Tag);

    P := PrevSelectedTasks.IndexOf(TaskEntry.Name);
    if P >= 0 then begin
      if TaskButton is TCheckBox then
        TCheckBox(TaskButton).Checked := not(toCheckedOnce in TaskEntry.Options)
      else
        TGroupRadioButton(TaskButton).Checked := not(toCheckedOnce in TaskEntry.Options);
      Continue;
    end;

    P := PrevDeselectedTasks.IndexOf(TaskEntry.Name);
    if P >= 0 then begin
      if TaskButton is TCheckBox then
        TCheckBox(TaskButton).Checked := False
      else
        TGroupRadioButton(TaskButton).Checked := False;
    end;
  end;
end;

function TWizardForm.GetSetupType(): PSetupTypeEntry;
var
  Index: Integer;
begin
  Index := TypesCombo.ItemIndex;
  if Index <> -1 then
    Result := PSetupTypeEntry(TypesCombo.Items.Objects[TypesCombo.ItemIndex])
  else
    Result := nil;
end;

procedure TWizardForm.SetSelectedComponents(const SelectedComponents, DeselectedComponents: TStringList);
var
  I, P: Integer;
  ComponentEntry: PSetupComponentEntry;
begin
  for I := 0 to Entries[seComponent].Count-1 do begin
    ComponentEntry := PSetupComponentEntry(Entries[seComponent][I]);

    if SelectedComponents <> nil then begin
      P := SelectedComponents.IndexOf(ComponentEntry.Name);
      if P >= 0 then begin
        ComponentsList.Checked[I] := True;
        Continue;
      end;
    end;

    if DeselectedComponents <> nil then begin
      P := DeselectedComponents.IndexOf(ComponentEntry.Name);
      if P >= 0 then
        ComponentsList.Checked[I] := False
    end;
  end;
end;

procedure TWizardForm.SetSelectedComponentsFromType(const TypeName: String; OnlySelectFixedComponents: Boolean);
var
  ComponentTypes: TStringList;
  ComponentEntry: PSetupComponentEntry;
  I: Integer;
begin
  ComponentTypes := TStringList.Create();
  for I := 0 to Entries[seComponent].Count-1 do begin
    ComponentEntry := PSetupComponentEntry(Entries[seComponent][I]);
    if not OnlySelectFixedComponents or (coFixed in ComponentEntry.Options) then begin
      ComponentTypes.CommaText := ComponentEntry.Types;
      ComponentsList.Checked[I] := ComponentTypes.IndexOf(TypeName) <> -1;
    end;
  end;
  ComponentTypes.Free();
end;

procedure TWizardForm.UpdateSelectTasksPage;
var
  SelectedComponents: TStringList;
begin
  SelectedComponents := TStringList.Create();
  GetSelectedComponents(SelectedComponents, False);
  CreateTaskButtons(SelectedComponents);
  SelectedComponents.Free();
end;

procedure TWizardForm.GetSelectedComponents(Components: TStringList; const Descriptions: Boolean);

  function GetString(ComponentEntry: PSetupComponentEntry; Descriptions: Boolean): String;
  begin
    if Descriptions then
      Result := ChangeDirConst(ComponentEntry.Description)
    else
      Result := ComponentEntry.Name;
  end;

var
  ComponentEntry: PSetupComponentEntry;
  I: Integer;
begin
  Components.Clear();
  for I := 0 to ComponentsList.Items.Count-1 do begin
    if ComponentsList.Checked[I] then begin
      ComponentEntry := PSetupComponentEntry(ComponentsList.ItemObject[I]);
      Components.Add(GetString(ComponentEntry, Descriptions));
    end;
  end;
end;

procedure TWizardForm.GetSelectedTasks(Tasks: TStringList; const Descriptions: Boolean);

  function GetString(TaskEntry: PSetupTaskEntry; Descriptions: Boolean): String;
  var
    Description: String;
    I, N: Integer;
  begin
    if Descriptions then begin
      Description := ChangeDirConst(TaskEntry.Description);
      N := Length(Description);
      Result := '';
      for I := 1 to N do begin
        if (I = N) or not ((Description[I] = '&') and (Description[I+1] <> '&')) then
          Result := Result + Description[I];
      end;
    end else
      Result := TaskEntry.Name;
  end;

var
  TaskEntry: PSetupTaskEntry;
  TaskButton: TButtonControl;
  Checked: Boolean;
  I: Integer;
begin
  Tasks.Clear();
  for I := 0 to TaskButtons.Count-1 do begin
    TaskButton := TaskButtons[I];
    if TaskButton is TCheckBox then
      Checked := TCheckBox(TaskButtons[I]).Checked
    else
      Checked := TGroupRadioButton(TaskButtons[I]).Checked;
    if Checked then begin
      TaskEntry := PSetupTaskEntry(TaskButton.Tag);
      Tasks.Add(GetString(TaskEntry, Descriptions));
    end;
  end;
end;

function TWizardForm.GetFirstTaskButton(): TButtonControl;
var
  TaskButton: TButtonControl;
  I: Integer;
begin
  for I := 0 to TaskButtons.Count-1 do begin
    TaskButton := TaskButtons[I];
    if not(TaskButton is TGroupRadioButton) or TGroupRadioButton(TaskButton).Checked then begin
      Result := TaskButton;
      Exit;
    end;
  end;
  Result := nil;
end;

procedure EnumProc (const Filename: String; Param: Pointer);
begin
  TStringList(Param).Add (AnsiLowerCaseFileName(Filename));
end;

function TWizardForm.PrepareToInstall: Boolean;

  function PreviousInstallNotCompleted: Boolean;
  var
    SL: TStringList;

    function CheckForFile(Filename: String): Boolean;
    var
      J: Integer;
    begin
      if UsingWinNT then
        Filename := AnsiLowerCaseFileName(Filename)
      else
        Filename := AnsiLowerCaseFileName(GetShortName(Filename));
      for J := 0 to SL.Count-1 do begin
        if SL[J] = Filename then begin
          Result := True;
          Exit;
        end;
      end;
      Result := False;
    end;

  var
    I: Integer;
    CurFile: PSetupFileEntry;
    H: THandle;
    FindData: TWin32FindData;
    D: String;
  begin
    Result := False;
    if Entries[seFile].Count = 0 then
      Exit;
    SL := TStringList.Create;
    try
      EnumFileReplaceOperationsFilenames (EnumProc, SL);
      if SL.Count = 0 then
        Exit;
      for I := 0 to Entries[seFile].Count-1 do begin
        CurFile := PSetupFileEntry(Entries[seFile][I]);
        if (CurFile^.FileType = ftUserFile) and
           ShouldProcessEntry(WizardComponents, WizardTasks, CurFile.Components, CurFile.Tasks) then begin
          if CurFile^.LocationEntry <> -1 then begin
            { Non-external file }
            if CheckForFile(ChangeDirConst(CurFile^.DestName)) then begin
              Result := True;
              Exit;
            end;
          end
          else begin
            { External file }
            H := FindFirstFile(PChar(ChangeDirConst(CurFile^.SourceFilename)),
              FindData);
            if H <> INVALID_HANDLE_VALUE then begin
              try
                repeat
                  if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY = 0 then begin
                    D := ChangeDirConst(CurFile^.DestName);
                    if not(foCustomDestName in CurFile^.Options) then
                      D := D + FindData.cFileName;
                    if CheckForFile(D) then begin
                      Result := True;
                      Exit;
                    end;
                  end;
                until not FindNextFile(H, FindData);
              finally
                Windows.FindClose (H);
              end;
            end;
          end;
        end;
      end;
    finally
      SL.Free;
    end;
  end;

begin
  Result := True;
  PreparingErrorPaintBox.Visible := False;
  PreparingLabel.Visible := False;
  Update;
  if PreviousInstallNotCompleted then begin
    PreparingLabel.Caption := ExpandSetupMessage(msgPreviousInstallNotCompleted) +
      SNewLine + SNewLine + SNewLine + SetupMessages[msgCannotContinue];
    AdjustLabelHeight (PreparingLabel);
    PreparingErrorPaintBox.Visible := True;
    PreparingLabel.Visible := True;
    Result := False;
    Exit;
  end;
end;

procedure TWizardForm.CurPageChanged;
{ Call this whenever the current page is changed }
var
  FirstTaskButton: TButtonControl;
  V: Boolean;
label 1;
begin
  Notebook1.PageIndex := NotebookPages[CurPage, 0];
  if NotebookPages[CurPage, 1] <> -1 then
    Notebook2.PageIndex := NotebookPages[CurPage, 1];

  { Set button visibility and captions }
  V := not(CurPage in [Low(TWizardPage), wpInstalling, wpInfoAfter, wpFinished]);
  if (SetupHeader.InfoAfterText <> '') and (CurPage = wpFinished) then
    V := True;

  BackButton.Visible := V;
  NextButton.Visible := CurPage <> wpInstalling;
  case CurPage of
    wpLicense: NextButton.Enabled := LicenseAcceptedRadio.Checked;
    wpPreparing: NextButton.Enabled := False;
  else
    NextButton.Enabled := True;
  end;
  CancelButton.Visible := CurPage < wpInfoAfter;
  CancelButton.Enabled := (CurPage <> wpInstalling) or not InitNoCancel;

  BackButton.Caption := SetupMessages[msgButtonBack];
  case CurPage of
    wpReady: begin
        NextButton.Caption := SetupMessages[msgButtonInstall];
        CancelButton.Caption := SetupMessages[msgButtonCancel];
      end;
    wpFinished: begin
        NextButton.Caption := SetupMessages[msgButtonFinish];
        CancelButton.Caption := SetupMessages[msgButtonCancel];
      end;
  else
  1:NextButton.Caption := SetupMessages[msgButtonNext];
    CancelButton.Caption := SetupMessages[msgButtonCancel];
  end;

  { Set the page description }
  if CurPage in [Low(PageCaptions)..High(PageCaptions)] then
    PageNameLabel.Caption := SetupMessages[PageCaptions[CurPage]];
  if CurPage in [Low(PageDescriptions)..High(PageDescriptions)] then
    PageDescriptionLabel.Caption := PageDescriptions[CurPage];
  if CurPage in [wpWelcome, wpFinished] then
    Notebook1.Color := clWindow
  else
    Notebook1.Color := clBtnFace;

  BeveledLabel.Visible := (SetupMessages[msgBeveledLabel] <> '') and
    not(CurPage in [wpWelcome, wpFinished]);

  { Adjust focus }
  case CurPage of
    wpWelcome: ActiveControl := NextButton;
    wpLicense: ActiveControl := LicenseMemo;
    wpPassword: ActiveControl := PasswordEdit;
    wpInfoBefore: ActiveControl := InfoBeforeMemo;
    wpUserInfo: ActiveControl := UserInfoNameEdit;
    wpSelectDir: ActiveControl := DirEdit;
    wpSelectComponents:  ActiveControl := TypesCombo;
    wpSelectProgramGroup: begin
        if not NoIconsCheck.Checked then
          ActiveControl := GroupEdit
        else
          ActiveControl := NoIconsCheck;
      end;
    wpSelectTasks: begin
        FirstTaskButton := GetFirstTaskButton();
        if FirstTaskButton <> nil then
          ActiveControl := FirstTaskButton
      end;
    wpReady: ActiveControl := NextButton;
    wpPreparing: ActiveControl := CancelButton;
    wpInstalling: begin
        if CancelButton.Enabled then
          ActiveControl := CancelButton
        else
          ActiveControl := nil;
      end;
    wpInfoAfter: ActiveControl := InfoAfterMemo;
    wpFinished: begin
        if (FirstRunButton <> nil) and FirstRunButton.Visible then
          ActiveControl := FirstRunButton
        else
        if YesRadio.Visible and YesRadio.Checked then
          ActiveControl := YesRadio
        else
        if NoRadio.Visible and NoRadio.Checked then
          ActiveControl := NoRadio
        else
          ActiveControl := NextButton;
      end;
  end;
end;

function TWizardForm.SkipCurPage: Boolean;

  function SkipSelectTasksPage: Boolean;
  var
    SelectedComponents: TStringList;
    I: Integer;
  begin
    //check if there is a task that has to be processed based on the selected
    //components
    SelectedComponents := TStringList.Create();
    GetSelectedComponents(SelectedComponents, False);
    Result := True;
    for I := 0 to Entries[seTask].Count-1 do begin
      with PSetupTaskEntry(Entries[seTask][I])^ do begin
        if ShouldProcessEntry(SelectedComponents, nil, Components, '') then begin
          Result := False;
          Break;
        end;
      end;
    end;
    SelectedComponents.Free();
  end;

begin
  Result :=
    ((CurPage = wpLicense) and (SetupHeader.LicenseText = '')) or
    ((CurPage = wpPassword) and not NeedPassword) or
    ((CurPage = wpInfoBefore) and (SetupHeader.InfoBeforeText = '')) or
    ((CurPage = wpUserInfo) and not(shUserInfoPage in SetupHeader.Options)) or
    ((CurPage = wpSelectDir) and ((shDisableDirPage in SetupHeader.Options) or not(shCreateAppDir in SetupHeader.Options))) or
    ((CurPage = wpSelectComponents) and not HasComponents) or
    ((CurPage = wpSelectProgramGroup) and ((shDisableProgramGroupPage in SetupHeader.Options) or not HasIcons)) or
    ((CurPage = wpSelectTasks) and SkipSelectTasksPage()) or
    ((CurPage = wpReady) and (shDisableReadyPage in SetupHeader.Options)) or
    ((CurPage = wpPreparing) and PrepareToInstall);
end;

procedure TWizardForm.NextButtonClick(Sender: TObject);

  function CheckPassword: Boolean;
  var
    S: String;
    SaveCursor: HCURSOR;
  begin
    S := PasswordEdit.Text;
    Result := GetCRC32(S[1], Length(S)) = SetupHeader.Password;
    if Result then begin
      NeedPassword := False;
      PasswordEdit.Text := '';
    end
    else begin
      { Delay for 750 ms when an incorrect password is entered to
        discourage brute-force attempts }
      SaveCursor := GetCursor;
      SetCursor (LoadCursor(0, IDC_WAIT));
      Sleep (750);
      SetCursor (SaveCursor);
      MsgBox (SetupMessages[msgIncorrectPassword], '', mbError, MB_OK);
      PasswordEdit.Text := '';
      PasswordEdit.SetFocus;
    end;
  end;

  function CheckUserInfoPage: Boolean;
  begin
    UserInfoNameEdit.Text := Trim(UserInfoNameEdit.Text);
    UserInfoOrgEdit.Text := Trim(UserInfoOrgEdit.Text);
    Result := UserInfoNameEdit.Text <> '';
    if Result then begin
      WizardUserInfoName := UserInfoNameEdit.Text;
      WizardUserInfoOrg := UserInfoOrgEdit.Text;
    end
    else begin
      MsgBox (SetupMessages[msgUserInfoNameRequired], '', mbError, MB_OK);
      UserInfoNameEdit.SetFocus;
    end;
  end;

  function Multiply32x32 (N1, N2: Cardinal): Cardinal;
  { Multiplies two 32-bit unsigned integers together. If the result is greater
    than $FFFFFFFF, the function returns $FFFFFFFF.
    Delphi versions < 4 don't natively support 64-bit multiplies, so we have to
    use assembler code to do this. }
  asm
    { EAX = N1, EDX = N2 }
    mul edx              { Multiplies EAX by EDX, places 64-bit result in EDX:EAX }
    test edx, edx        { EDX will be zero if result is < 4 GB }
    jz @@1
    mov eax, 0FFFFFFFFh  { If > 4 GB then cap at 4 GB }
    @@1:
  end;

  function GetTotalFreeSpace (const DriveRoot: String; var FreeSpace: Longint): Boolean;
  type
    TLargeIntegerRec = record
      Lo, Hi: Cardinal;
    end;
  var
    GetDiskFreeSpaceExFunc: function(lpDirectoryName: PAnsiChar;
      var lpFreeBytesAvailable, lpTotalNumberOfBytes: TLargeInteger;
      lpTotalNumberOfFreeBytes: PLargeInteger): BOOL; stdcall;
    FreeBytes, TotalBytes: TLargeInteger;
    SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters: Cardinal;
  begin
    { NOTE: The docs claim that GetDiskFreeSpace supports UNC paths on
      Windows 95 OSR2 and later. But that does not seem to be the case in my
      tests; it fails with error 50 on Windows 95 through Me.
      GetDiskFreeSpaceEx, however, *does* succeed with UNC paths, so use it
      if available. }
    GetDiskFreeSpaceExFunc := GetProcAddress(GetModuleHandle(kernel32),
      'GetDiskFreeSpaceExA');
    if Assigned(@GetDiskFreeSpaceExFunc) then begin
      Result := GetDiskFreeSpaceExFunc(PChar(DriveRoot), FreeBytes, TotalBytes, nil);
      if Result then begin
        if TLargeIntegerRec(FreeBytes).Hi = 0 then
          FreeSpace := TLargeIntegerRec(FreeBytes).Lo
        else
          Cardinal(FreeSpace) := Cardinal($FFFFFFFF);
      end;
    end
    else begin
      Result := GetDiskFreeSpace(PChar(DriveRoot), DWORD(SectorsPerCluster),
        DWORD(BytesPerSector), DWORD(FreeClusters), DWORD(TotalClusters));
      if Result then begin
        { Windows 95/98 cap the result of GetDiskFreeSpace at 2GB, but NT 4.0
          does not, so we must use a 64-bit multiply operation to avoid an
          overflow. }
        FreeSpace := Multiply32x32(BytesPerSector * SectorsPerCluster,
          FreeClusters);
      end;
    end;
  end;

  function SpaceString (const S: String): String;
  var
    I: Integer;
  begin
    Result := '';
    for I := 1 to Length(S) do begin
      if S[I] = ' ' then Continue;
      if Result <> '' then Result := Result + ' ';
      Result := Result + S[I];
    end;
  end;

  function CheckSelectDirPage: Boolean;
  const
    BadDirChars = '/:*?"<>|';
  var
    T: String;
    IsUNCPath: Boolean;
    I: Integer;
    TotalFreeSpace: Longint;
    RootPath: String;
  begin
    Result := False;

    { Remove any leading or trailing spaces }
    DirEdit.Text := Trim(DirEdit.Text);

    T := DirEdit.Text;  { reduce calls to GetText }

    { Check for UNC pathname }
    IsUNCPath := (Length(T) >= 2) and (T[1] = '\') and (T[2] = '\');
    if IsUNCPath and not(shAllowUNCPath in SetupHeader.Options) then begin
      MsgBox (SetupMessages[msgToUNCPathname], '', mbError, MB_OK);
      Exit;
    end;

    { Automatically change '\\' to '\' and remove any trailing backslash }
    if not IsUNCPath then
      T := RemoveExtraBackslashes(T)
    else
      T := '\\' + RemoveExtraBackslashes(Copy(T, 3, Maxint));
    T := RemoveBackslashUnlessRoot(T);
    DirEdit.Text := T;

    if not IsUNCPath then begin
      { Check if is at least 4 chars long and it has a drive letter, colon,
        and backslash }
      if not(shAllowRootDirectory in SetupHeader.Options) then
        I := 4
      else
        I := 3;
      if (Length(T) < I) or not(UpCase(T[1]) in ['A'..'Z']) or
         (T[2] <> ':') or (T[3] <> '\') then begin
        MsgBox (SetupMessages[msgInvalidPath], '', mbError, MB_OK);
        Exit;
      end;
    end
    else begin
      { Check if there's at least one backslash after the initial '\\' }
      if StrScan(@T[3], '\') = nil then begin
        MsgBox (SetupMessages[msgInvalidPath], '', mbError, MB_OK);
        Exit;
      end;
    end;

    { Check for invalid characters after 'x:' or '\\' }
    if LastDelimiter(BadDirChars, Copy(T, 3, Maxint)) <> 0 then begin
      MsgBox (FmtSetupMessage1(msgBadDirName32, SpaceString(BadDirChars)), '',
        mbError, MB_OK);
      Exit;
    end;

    { Check if it's a valid drive }
    RootPath := AddBackslash(ExtractFileDrive(T));
    if not DirExists(RootPath) then begin
      MsgBox (SetupMessages[msgInvalidDrive], '', mbError, MB_OK);
      Exit;
    end;

    { Check if enough disk space (using an unsigned long compare) }
    if GetTotalFreeSpace(RootPath, TotalFreeSpace) then begin
      if Cardinal(TotalFreeSpace) < Cardinal(MinimumSpace) then
        { If not, show warning }
        if MsgBox(FmtSetupMessage(msgDiskSpaceWarning,
             [IntToStr(MinimumSpace div 1024), IntToStr(TotalFreeSpace div 1024)]),
           SetupMessages[msgDiskSpaceWarningTitle],
           mbConfirmation, MB_YESNO or MB_DEFBUTTON2) <> ID_YES then
          Exit;
    end;

    { Check if directory already exists }
    if ((SetupHeader.DirExistsWarning = ddYes) or
        ((SetupHeader.DirExistsWarning = ddAuto) and (T <> PrevAppDir))) and
       DirExists(T) then
      { If so, ask if user wants to install there anyway }
      if MsgBox(FmtSetupMessage1(msgDirExists, T), SetupMessages[msgDirExistsTitle],
        mbConfirmation, MB_YESNO) <> ID_YES then Exit;

    { Check if directory *doesn't* already exist }
    if (shEnableDirDoesntExistWarning in SetupHeader.Options) and
       not DirExists(T) then
      { If not, ask if user wants to install there anyway }
      if MsgBox(FmtSetupMessage1(msgDirDoesntExist, T), SetupMessages[msgDirDoesntExistTitle],
        mbConfirmation, MB_YESNO) <> ID_YES then Exit;

    Result := True;
    WizardDirValue := T;
  end;

  function CheckSelectComponentsPage: Boolean;
  var
    ComponentEntry: PSetupComponentEntry;
    TotalFreeSpace: Longint;
    S: String;
    I: Integer;
  begin
    Result := False;

    if GetTotalFreeSpace(AddBackslash(ExtractFileDrive(DirEdit.Text)),
       TotalFreeSpace) then begin
      if Cardinal(TotalFreeSpace) < Cardinal(MinimumComponentsSpace) then
        if MsgBox(FmtSetupMessage(msgDiskSpaceWarning,
             [IntToStr(MinimumComponentsSpace div 1024), IntToStr(TotalFreeSpace div 1024)]),
           SetupMessages[msgDiskSpaceWarningTitle],
           mbConfirmation, MB_YESNO or MB_DEFBUTTON2) <> ID_YES then
          Exit;
    end;

    //now see if there are unchecked components that are already installed
    if PrevSelectedComponents.Count > 0 then begin
      S := '';
      for I := 0 to ComponentsList.Items.Count-1 do begin
        if not ComponentsList.Checked[I] then begin
          ComponentEntry := PSetupComponentEntry(ComponentsList.ItemObject[I]);
          if not (coDisableNoUninstallWarning in ComponentEntry.Options) then begin
            if PrevSelectedComponents.IndexOf(ComponentEntry.Name) <> -1 then begin
              if S <> '' then
                S := S + #13;
              S := S + ChangeDirConst(ComponentEntry.Description);
            end;
          end;
        end;
      end;

      if (S = '') or (MsgBox(FmtSetupMessage1(msgNoUninstallWarning, S), SetupMessages[msgNoUninstallWarningTitle], mbConfirmation, MB_YESNO) = ID_YES) then
        Result := True;
    end else
      Result := True;
  end;

  function CheckSelectProgramGroupPage: Boolean;
  const
    BadGroupChars = '/:*?"<>|,';
  var
    T: String;
  begin
    Result := False;

    if not NoIconsCheck.Checked then begin
      { Remove any leading or trailing spaces and out of place backslashes }
      T := RemoveExtraBackslashes(Trim(GroupEdit.Text));
      if (T <> '') and (T[1] = '\') then
        Delete (T, 1, 1);
      T := RemoveBackslash(T);
      GroupEdit.Text := T;

      if T = '' then begin
        MsgBox (SetupMessages[msgMustEnterGroupName], '', mbError, MB_OK);
        Exit;
      end;

      { Check for invalid characters }
      if LastDelimiter(BadGroupChars, T) <> 0 then begin
        MsgBox (FmtSetupMessage1(msgBadGroupName, SpaceString(BadGroupChars)),
          '', mbError, MB_OK);
        Exit;
      end;
    end;

    Result := True;
  end;

  procedure UpdateReadyPage;
  var
    TypeEntry: PSetupTypeEntry;
    SelectedComponents, SelectedTasks: TStringList;
    Space, S: String;
    I: Integer;
  begin
    Space := Format('%6s', ['']);

    ReadyMemo.Visible := False;
    if not (shDisableReadyMemo in SetupHeader.Options) then begin
      ReadyMemo.Lines.Clear();

      if shUserInfoPage in SetupHeader.Options then begin
        ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoUserInfo]);
        ReadyMemo.Lines.Append(Space+UserInfoNameEdit.Text);
        if UserInfoOrgEdit.Text <> '' then
          ReadyMemo.Lines.Append(Space+UserInfoOrgEdit.Text);
      end;

      if (shAlwaysShowDirOnReadyPage in SetupHeader.Options) or
         (not(shDisableDirPage in SetupHeader.Options) and
          (shCreateAppDir in SetupHeader.Options)) then begin
        if ReadyMemo.Lines.Count > 0 then
          ReadyMemo.Lines.Append('');
        ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoDir]);
        ReadyMemo.Lines.Append(Space+DirEdit.Text);
      end;

      if HasComponents then begin
        if ReadyMemo.Lines.Count > 0 then
          ReadyMemo.Lines.Append('');
        ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoType]);
        TypeEntry := PSetupTypeEntry(TypesCombo.Items.Objects[TypesCombo.ItemIndex]);
        ReadyMemo.Lines.Append(Space+ChangeDirConst(TypeEntry.Description));

        SelectedComponents := TStringList.Create();
        GetSelectedComponents(SelectedComponents, True);
        if SelectedComponents.Count > 0 then begin
          if ReadyMemo.Lines.Count > 0 then
            ReadyMemo.Lines.Append('');
          ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoComponents]);
          for I := 0 to SelectedComponents.Count-1 do
            ReadyMemo.Lines.Append(Space+SelectedComponents[I]);
        end;
        SelectedComponents.Free();
      end;

      if HasIcons and not NoIconsCheck.Checked and
         ((shAlwaysShowGroupOnReadyPage in SetupHeader.Options) or
          not(shDisableProgramGroupPage in SetupHeader.Options)) then begin
        if ReadyMemo.Lines.Count > 0 then
          ReadyMemo.Lines.Append('');
        ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoGroup]);
        ReadyMemo.Lines.Append(Space+GroupEdit.Text);
      end;

      SelectedTasks := TStringList.Create();
      GetSelectedTasks(SelectedTasks, True);
      if SelectedTasks.Count > 0 then begin
        if ReadyMemo.Lines.Count > 0 then
          ReadyMemo.Lines.Append('');
        ReadyMemo.Lines.Append(SetupMessages[msgReadyMemoTasks]);
        for I := 0 to SelectedTasks.Count-1 do
          ReadyMemo.Lines.Append(Space+SelectedTasks[I]);
      end;
      SelectedTasks.Free();

      ReadyMemo.SelStart := 0;
      ReadyMemo.SelLength := 0;
    end;

    if ReadyMemo.Lines.Count > 0 then begin
      S := SetupMessages[msgReadyLabel2a];
      ChangeReadyLabel(S);
      ReadyMemo.Visible := True;
    end else begin
      S := SetupMessages[msgReadyLabel2b];
      ChangeReadyLabel(S);
    end;
  end;

begin
  case CurPage of
    wpLicense: if not LicenseAcceptedRadio.Checked then Exit;  { paranoia }
    wpPassword: if not CheckPassword then Exit;
    wpUserInfo: if not CheckUserInfoPage then Exit;
    wpSelectDir: if not CheckSelectDirPage then Exit;
    wpSelectComponents: if not CheckSelectComponentsPage then Exit;
    wpSelectProgramGroup: if not CheckSelectProgramGroupPage then Exit;
  end;

  { Go to the next page, or close wizard if it was on the last page }
  repeat
    if CurPage = wpPreparing then begin
      CurPage := wpInstalling;
      CurPageChanged;
      PostMessage (MainForm.Handle, WM_SETUPNEXTSTEP, 0, 0);
      Exit;
    end else if CurPage = wpFinished then begin
      { Note: Hide isn't used here because on Delphi 3+ it makes the taskbar
        button disappear if this was the only Visible form. }
      MainForm.HideWizard := True;
      MainForm.UpdateWizardFormVisibility;
      PostMessage (MainForm.Handle, WM_SETUPNEXTSTEP, 0, 0);
      Exit;
    end;

    Inc (CurPage);
    //even if we're skipping a page, we should still update it
    case CurPage of
      wpSelectTasks: begin
          if NeedSelectTasksPageUpdate then begin
            UpdateSelectTasksPage();
            NeedSelectTasksPageUpdate := False;
          end;
        end;
      wpReady: UpdateReadyPage;
    end;
  until not SkipCurPage;

  CurPageChanged;
end;

procedure TWizardForm.BackButtonClick(Sender: TObject);
begin
  if CurPage = Low(TWizardPage) then Exit;

  { Go to the previous page }
  Dec (CurPage);
  while SkipCurPage do
    Dec (CurPage);
  CurPageChanged;
end;

procedure TWizardForm.CancelButtonClick(Sender: TObject);
begin
  { Clicking Cancel will do the same thing as the Close button }
  Close;
end;

procedure TWizardForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
begin
  { Redirect an attempt to close this form to MainForm }
  MainForm.Close;
  CanClose := False;
end;

procedure TWizardForm.ReadGroupList;
{ Fills in the list box of existing groups }
var
  Path, StartupName: String;
  H: THandle;
  FindData: TWin32FindData;
  S: String;
begin
  try
    Path := AddBackslash(ExpandFileName(GetShellFolder(
      not(shAlwaysUsePersonalGroup in SetupHeader.Options), sfPrograms, True)));
    if Path = '' then Exit;

    { Determine the name of the Startup folder so that we can hide it from the
      list }
    StartupName := ExpandFileName(GetShellFolder(
      not(shAlwaysUsePersonalGroup in SetupHeader.Options), sfStartup, True));
    if StartupName <> '' then begin
      if AnsiCompareFileName(Path, ExtractFilePath(StartupName)) = 0 then
        StartupName := ExtractFileName(StartupName)
      else
        StartupName := '';
    end;

    H := FindFirstFile(PChar(Path + '*'), FindData);
    if H <> INVALID_HANDLE_VALUE then begin
      repeat
        if FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY <> 0 then begin
          S := FindData.cFileName;
          if (S <> '.') and (S <> '..') and
             (AnsiCompareFileName(S, StartupName) <> 0) then
            GroupList.Items.Add (S);
        end;
      until not FindNextFile(H, FindData);
      Windows.FindClose (H);
    end;
  except
    { ignore any exceptions }
  end;
end;

procedure TWizardForm.DirListChange(Sender: TObject);
begin
  if not(shDisableAppendDir in SetupHeader.Options) then
    DirEdit.Text := AddBackslash(DirList.Directory) +
      ExtractFileName(ChangeDirConst(SetupHeader.DefaultDirName))
  else
    DirEdit.Text := DirList.Directory;
end;

procedure TWizardForm.TypesComboChange(Sender: TObject);
var
  TypeEntry: PSetupTypeEntry;
begin
  //select the components for this type. if the type is custom only select
  //fixed components
  TypeEntry := PSetupTypeEntry(TypesCombo.Items.Objects[TypesCombo.ItemIndex]);
  SetSelectedComponentsFromType(TypeEntry.Name, (toIsCustom in TypeEntry.Options));

  //if customization is possible remember the type and components that are
  //selected, so that we can reselect the setup type later if after customization
  //the user didn't really change anything
  //also hide the components list if necessary
  if HasCustomType then begin
    InitialSetupTypeIndex := TypesCombo.ItemIndex;
    GetSelectedComponents(InitialSelectedComponents, False);
    if not (shAlwaysShowComponentsList in SetupHeader.Options) then begin
      ComponentsList.Visible := toIsCustom in TypeEntry.Options;
      ComponentsDiskSpaceLabel.Visible := ComponentsList.Visible;
    end;
  end;

  CalcComponentsSpace();
  NeedSelectTasksPageUpdate := True;
end;

procedure TWizardForm.ComponentsListClickCheck(Sender: TObject);
var
  SelectedComponents: TStringList;
  TypeEntry: PSetupTypeEntry;
  Equals: Boolean;
  I: Integer;
begin
  //first see if this current selection equals the initial selection
  //if so, reselect the initial setup type
  SelectedComponents := TStringList.Create();
  GetSelectedComponents(SelectedComponents, False);
  Equals := SelectedComponents.Equals(InitialSelectedComponents);
  SelectedComponents.Free();

  if Equals then begin
    //select the intial type
    TypesCombo.ItemIndex := InitialSetupTypeIndex;
  end else begin
    //select a custom type
    for I := 0 to Entries[seType].Count-1 do begin
      TypeEntry := Entries[seType][I];
      if (toIsCustom in TypeEntry.Options) then begin
        TypesCombo.ItemIndex := TypesCombo.Items.IndexOfObject(TObject(TypeEntry));
        SetSelectedComponentsFromType(TypeEntry.Name, True);
        Break;
      end;
    end
  end;

  CalcComponentsSpace();
  NeedSelectTasksPageUpdate := True;
end;

procedure TWizardForm.GroupEditChange(Sender: TObject);
var
  T: String;
  I: Integer;
  SaveOnClick: TNotifyEvent;
label 1;
begin
  T := GroupEdit.Text;
  with GroupList do begin
    for I := 0 to Items.Count-1 do
      if AnsiCompareFileName(T, Items[I]) = 0 then
        goto 1;
    I := -1;
  1:if ItemIndex <> I then begin
      SaveOnClick := OnClick;
      { Temporarily disabled the OnChange event }
      OnClick := nil;
      try
        ItemIndex := I;
      finally
        OnClick := SaveOnClick;
      end;
    end;
  end;
end;

procedure TWizardForm.GroupListClick(Sender: TObject);
var
  SaveOnChange: TNotifyEvent;
begin
  with GroupList do
    if ItemIndex <> -1 then begin
      SaveOnChange := GroupEdit.OnChange;
      { Temporarily disabled the OnChange event }
      GroupEdit.OnChange := nil;
      try
        GroupEdit.Text := Items[ItemIndex];
      finally
        GroupEdit.OnChange := SaveOnChange;
      end;
    end;
end;

procedure TWizardForm.NoIconsCheckClick(Sender: TObject);
const
  ColorChange: array[Boolean] of TColor = (clBtnFace, clWindow);
begin
  with GroupEdit do begin
    Enabled := not NoIconsCheck.Checked;
    Color := ColorChange[Enabled];
  end;
  with GroupList do begin
    Enabled := not NoIconsCheck.Checked;
    Color := ColorChange[Enabled];
    if not Enabled then ItemIndex := -1;
  end;
end;

procedure TWizardForm.ReadyMemoKeyPress(Sender: TObject; var Key: Char);
begin
  if Key = #13 then
    NextButton.Click();
end;

procedure TWizardForm.PicturePaintBoxPaint(Sender: TObject);
var
  Bitmap: TBitmap;
  BackColor: TColor;
  R: TRect;
  X, Y: Integer;
begin
  if Sender = SmallPicturePaintBox then begin
    Bitmap := WizardSmallImage;
    BackColor := SetupHeader.WizardSmallImageBackColor;
  end else begin
    Bitmap := WizardImage;
    BackColor := SetupHeader.WizardImageBackColor;
  end;

  with TPaintBox(Sender) do begin
    R := ClientRect;
    { Draw the background }
    Canvas.Brush.Color := BackColor;
    Canvas.FillRect (R);
    { And finally draw the image }
    if Assigned(Bitmap) then begin
      X := R.Left + ((R.Right - R.Left) - Bitmap.Width) div 2;
      if X < 0 then X := 0;
      Y := R.Top + ((R.Bottom - R.Top) - Bitmap.Height) div 2;
      if Y < 0 then Y := 0;
      Canvas.Draw (X, Y, Bitmap);
    end;
  end;
end;

procedure TWizardForm.WMSysCommand (var Message: TWMSysCommand);
begin
  if Message.CmdType and $FFF0 = SC_MINIMIZE then
    { A minimize button is shown on the wizard form when (shWindowVisible in
      SetupHeader.Options). When it is clicked we want to minimize the whole
      application. }
    Application.Minimize
  else
  if Message.CmdType = 9999 then
    MainForm.ShowAboutBox
  else
    inherited;
end;

procedure TWizardForm.LicenseAcceptedRadioClick(Sender: TObject);
begin
  if CurPage = wpLicense then
    NextButton.Enabled := True;
end;

procedure TWizardForm.LicenseNotAcceptedRadioClick(Sender: TObject);
begin
  if CurPage = wpLicense then
    NextButton.Enabled := False;
end;

procedure TWizardForm.PreparingErrorPaintBoxPaint(Sender: TObject);
var
  Bmp: TBitmap;
  R: TRect;
begin
  Bmp := TBitmap.Create;
  try
    Bmp.Handle := LoadBitmap(HInstance, 'STOPIMAGE');
    R := Rect(0, 0, 16, 16);
    PreparingErrorPaintBox.Canvas.BrushCopy (R, Bmp, R, clSilver);
  finally
    Bmp.Free;
  end;
end;

end.
